/**
 * PrimeBuilder.js - Taking control of content editable
 * @constructor
 * @param {Object} [userSettings] user options
 */
var PrimeBuilder = function (selector, userOptions) {
	"use strict";

	var prime = this,
		defaultOptions = Object.assign({}, PrimeBuilder.defaultOptions),
		options = this.options = Object.assign(defaultOptions, userOptions),
		html = PrimeBuilder.Html,
		utils = PrimeBuilder.Utils,
		elements = PrimeBuilder.Elements,
		settings = PrimeBuilder.Settings,
		cache = new PrimeBuilder.Cache(),
		action = new PrimeBuilder.Action(this),
		apply = new PrimeBuilder.Apply(this),
		picker = new PrimeBuilder.Picker(this);

	this.variants = ib2variants;
	this.currentVariation = $('#ib2-current-variation').val();
	this.globals = ib2options;
	this.selector = selector;
	this.options  = options;
	this.html = html;
	this.utils = utils;
	this.elements = elements;
	this.cache = cache;
	this.gui = null;
	this.action = action;
	this.interact = null;
	this.apply = apply;
	this.picker = picker;
	this.settings = settings;
	this.fa = window.FontAwesome;
	this.pagewidth = ib2globalsettings.settings.width || 1200;
	this.iframe = null;
	this.frameDoc = null;
	this.frameHtml = null;
	this.frameBody = null;
	this.history = false;
	this.permalink = $('#ib2-permalink').val();
	this.faIcons = ib2icons;
	this.postStatus = ib2post['status'] || 'draft';
	this.postData = ib2post;
	this.gatewayUrl = false;
	this.dirty = false;
	this.savingHistory = false;
	this.restored = false;
	this.sortableReload = false;
	this.lastautosave = Math.floor(Date.now()/1000);
	this.tempShortcode = {};
	this.esplists = ib2lists;
	this.timeoutHandler = null;

	if ( this.postStatus == 'true' ) {
		this.postStatus = 'publish';
	}

	// global variable
	window.primeObj = prime;
	
	picker.setup(); // setup color picker

	// convert old page data format
	$(window).load(function(){
		prime.init();
	});

	this.restoreHistory();
};

PrimeBuilder.prototype = {
	isHtmlElement: function (obj) {
   		return ( typeof obj === "object" ) && (obj !== null) &&
   			(obj !== false) && ( obj.nodeType );
	},

	decodeHTML: function (html) {
		var txt = document.createElement('textarea');
		txt.innerHTML = html;
		return txt.value;
	},

	init: function() {
		var self = this;
		$.ajax({
			type: "GET",
			url: ajaxurl,
			data: {
				action: 'ib2_load_editor',
				post_id: $('#ib2-post-id').val(),
				variant_id: $('#ib2-variant-id').val()
			},
			success: function(response) {
				var is_old = false,
					isAutosave = self.isAutosave();

				if ( response == 'old' ) {
					is_old = true;
				}

				if ( isAutosave && response.hasOwnProperty('html') ) {
					is_old = (!response.layers && !response.elements)
				}

				self.iframe = $('iframe#ib2-editor-content').get(0);

				window.FrameWindow = self.iframe.contentWindow;
				window.FrameDocument = self.iframe.contentWindow.document;

				self.frameDoc = $(window.FrameDocument);
				self.frameHtml = $(window.FrameDocument).find("html");
				self.frameBody = $(window.FrameDocument).find("body");

				self.interact = new PrimeBuilder.Interact(self);
				self.gui = new PrimeBuilder.Gui(self);

				self.interact.setup(); // setup interaction
				self.gui.setup(); // setup GUI

				if ( isAutosave && response.hasOwnProperty('html') ) {
					ib2globalsettings = response.data;
					if ( ib2globalsettings.hasOwnProperty('cssrules') ) {
						try {
							ib2cssrules = ib2globalsettings.cssrules;
						} catch(e) {
							// do nothing
						}
					}

					if ( ib2globalsettings.hasOwnProperty('videos') ) {
						self.cache.videos = ib2globalsettings.videos;
					}
					self.pagewidth = ib2globalsettings.settings.width || 1200;
					self.frameBody.find('#screen-container').html(response.html);
				}

				if ( typeof ib2cssrules !== "undefined" ) {
					self.cache.cssRules = ib2cssrules;
				}

				if ( ib2globalsettings.hasOwnProperty('videos') ) {
					self.cache.videos = ib2globalsettings.videos;
				}

				if ( is_old ) {
					// update old layout
					self.updateLayout();
				}

				// fix unrecognized text element
				self.updateUnrecognizedText();

				// load element icons
				self.loadElementIcons();

				// load tooltip
				//$('[data-toggle="tooltip"]').tooltip();

				// update questions
				self.updateQuestions();

				if ( is_old ) {
					self.saveLayers();
					self.objectify();
				} else {
					self.cache.layers = response.layers;
					self.cache.contents = response.elements;
				}

				if ( isAutosave && response.hasOwnProperty('html') && self.cache.cssRules ) {
					self.gui.buildCss(self.cache.cssRules);
				}

				if ( !is_old ) {
					// render countdowns, shortcodes & comments
					self.runCountdowns();
					self.renderElements();
				}

				// global fix and update layout
				self.globalRender(self);

				// fetch images from instabuilder server
				self.fetchImages();

				self.timeoutHandler = setTimeout(function(){
					// init drag and drop
					self.sortable();

					// init history for undo/redo
					$('body').trigger('initHistory');

					// save first history
					self.saveHistory();

					// disable undo
					$('#undo-btn').addClass('disabled');

					clearTimeout(self.timeoutHandler);
				}, 500);
			}
		});
	},

	restoreHistory: function() {
		var self = this,
			cache = this.cache;

		$('body').on('restore-history', function(e, data){
			if ( data ) {
				var is_old = (!data.layers && !data.elements),
					page_data = data.data;

				// save history
				self.saveHistory();
				self.savingHistory = true;
				setTimeout(function(){
					self.savingHistory = false;
				}, 2000);

				self.disableSortable();
				self.frameBody.find('#screen-container').html(data.html);
				ib2globalsettings = page_data;
				if ( page_data.hasOwnProperty('cssrules') ) {
					ib2cssrules = page_data.cssrules;
					self.cache.cssRules = ib2cssrules;
				}

				if ( page_data.hasOwnProperty('videos') ) {
					self.cache.videos = page_data.videos;
				}

				// fix the old wide background box
				// must be placed here to make sure
				// we can update the background video and image
				self.fixWideBox(true);

				self.pagewidth = ib2globalsettings.settings.width || 1200;
				if ( is_old ) {
					// update old layout
					self.updateLayout();
				}

				// fix unrecognized text element
				self.updateUnrecognizedText();

				// init drag and drop
				self.sortable();

				// update questions
				self.updateQuestions();

				if ( is_old ) {
					self.saveLayers();
					self.objectify();
				} else {
					self.cache.layers = data.layers;
					self.cache.contents = data.elements;
					// render countdowns, shortcodes & comments
					self.runCountdowns();
					self.renderElements();
				}

				if ( self.cache.cssRules ) {
					self.gui.buildCss(self.cache.cssRules);
				}

				self.globalRender(self);
			}
		});
	},

	updateLayout: function() {
		// update sections and columns
		this.updateColumns();
		this.updateSections(); 

		// mark all elements with new class
		this.updateElements();

		// update old elements
		this.updateBoxes(); // update box
		this.updateImages(); // update image
		this.updateVideos(); // update videos
		this.updateButtons(); // update buttons
		this.updateCountdowns(); // update countdown
		this.updateForms(); // update form
		this.updateMiscElements(); // update some elements
		this.updateBackgrounds(); // update background images
		this.getCustomStyles(); // get custom css

		// mark all elements with selection class
		this.updateForSelection();
	},

	updateUnrecognizedText: function() {
		if ( this.frameBody.find('[data-el="text"]').length ) {
			this.frameBody.find('[data-el="text"]').each(function(){
				var el = $(this);

				if ( !el.hasClass('ib2-text-el') ) {
					el.addClass('ib2-text-el');
				}

				if ( !el.hasClass('is-element') ) {
					el.addClass('is-element');
				}
			});
		}
	},

	fetchImages: function() {
		var images = [];
		// update image URL
		this.frameBody.find('.ib2-image-el').each(function(){
			var $this = $(this),
				elID = $this.attr('id'),
				img = $this.find('img'),
				src = img.attr('src');

			if ( src.indexOf('static.instabuilder.com') > -1 ) {
				images.push(src);
				$.ajax({
					type: "POST",
					url: ajaxurl,
					data: {
						action: 'ib2_download_image',
						img_url: 'https:' + src,
						post_id: $('#ib2-post-id').val(),
						is_fix: true,
					},
					success: function(response) {
						if ( response.success ) {
							$this.find('source').attr('srcset', response.img);
							img.attr('src', response.img);
						}
						$('#editor-loader').addClass('hidden');
						$('#editor-container').removeClass('hidden');
					}
				});
			}
		});

		// if there's no images to fetch
		if ( images.length <= 0 ) {
			$('#editor-loader').addClass('hidden');
			$('#editor-container').removeClass('hidden');
		}
	},

	globalRender: function(self) {
		var cache = self.cache;

		// set status
		if ( self.postStatus == 'publish' ) {
			$('.publish-page').hide();
		}

		// remove clearfix
		self.frameBody.find('.el-cols').each(function(){
			$(this).find(' > .clearfix').remove();
		});

		// fix image url
		self.frameBody.find('.ib2-image-el').each(function(){
			var $img = $(this),
				src = $img.find('img').attr('src');

			src = src.replace(/^(https?|ftp):/, '');
			$img.find('img').attr('src', src);
			$img.find('source').attr('srcset', src);
			var middleWidth = parseInt($img.find('> .el-content').css('width').replace(/px/gi, ''));
			if ( middleWidth >= self.pagewidth ) {
				var paddingRight = parseInt($img.closest('.el-content').css('paddingRight').replace(/px/gi, '')),
					correctWidth = middleWidth - (paddingRight*2);

				$img.find('> .el-content').css('width', correctWidth + 'px');
			}
		});

		// set permalink
		var letter = $('#ib2-current-variation').val();
		var preview_url = ( self.permalink.indexOf('?') > -1 ) ? self.permalink + '&preview=true&variant=' + letter : self.permalink + '?preview=true&variant=' + letter;
		$('.view-page-btn').attr('href', preview_url);

		// update today's date
		if ( self.frameBody.find('#screen-container').find('.ib2-date-el').length ) {
			self.frameBody.find('#screen-container').find('.ib2-date-el').each(function(){
				var el = $(this),
					offset = el.attr('data-tz'),
					format = el.attr('data-format');

				el.html(moment().tz(offset).format(format));
			});
		}

		// remove is-selected class
		self.frameBody.find('.is-selected').removeClass('is-selected');

		var pageSettings = cache.contents['page']['screen-container']['settings']['background'];
		if ( pageSettings ) {
			// fix global image background
			// check if for background placeholder if exists
			if ( !self.frameBody.find('#screen-container-box-img').length ) {
				// check if target has background
				var bgImg = pageSettings['backgroundImage'];
				if ( bgImg !== 'none' ) {
					var bgAttach = pageSettings['backgroundAttachment'],
						bgPos = bgAttach == 'scroll' ? pageSettings['backgroundPosition'] : 'center top',
						bgTile = pageSettings['backgroundRepeat'],
						bgSize = pageSettings['backgroundSize'] || 'auto';

					// target must not have background image
					self.frameBody.find('#screen-container').css('backgroundImage', 'none');

					// create new background image placeholder
					self.frameBody.find('#screen-container').prepend('<div id="screen-container-box-img" class="is-background"></div>');

					bgImg = bgImg.replace(new RegExp('{%IB2_PLUGIN_DIR%}', 'g'), $('#ib2-plugin-url').val());
					var css = {
						backgroundColor: 'transparent',
						backgroundImage: 'url("' + bgImg + '")',
						backgroundRepeat: bgTile,
						backgroundPos: bgPos,
						backgroundSize: bgSize,
						backgroundAttachment: bgAttach,
						position: 'absolute',
						top: 0,
						left: 0,
						maxWidth: '100%',
						width: '100%',
						height: self.frameBody.find('#screen-container').css('height')
					};

					// apply background image
					self.frameBody.find('#screen-container-box-img').css(css);
				}
			}
		}

		// set correct padding if notification bar is shown
		if ( self.frameBody.find('#screen-container').find('.ib2-notification-bar').length ) {
			var barHeight = self.frameBody.find('#screen-container').find('.ib2-notification-bar > .ib2-notification-text').outerHeight();
			self.frameBody.find('#screen-container').css('padding-top', barHeight + 'px');
		}

		self.frameBody.find('.ib2-menu-el').each(function(){
			// fix missing hamburger
			var burger = '<button type="button" class="ib2-navbar-toggle">\
						<span class="sr-only">Toggle navigation</span>\
						<span class="icon-bar"></span>\
						<span class="icon-bar"></span>\
						<span class="icon-bar"></span>\
					</button>';

			var $this = $(this);

			if ( !$this.find('nav').length ) {
				$this.find('.ib2-navi').wrap('<nav class="clearfix"></nav>');
			}

			if ( !$this.find('.ib2-navbar-toggle').length ) {
				$this.prepend(burger);
			}
		});

		// fix opacity and background border
		self.frameBody.find('.ib2-element').each(function(){
			var $this = $(this),
				type = $this.attr('data-el'),
				elID = $this.attr('id'),
				target;

			if ( type == 'section' || type == 'wsection' || type == 'columns' || type == 'box' || type == 'wbox' ) {
				target = ( type == 'box' ) ? self.frameBody.find('#' + elID + '-box') : self.frameBody.find('#' + elID + ' > .el-content');

				// fix opacity
				var opac = target.css('opacity'),
					backgroundColor = target.css('backgroundColor');

				if ( opac < 1 ) {
					target.css('opacity', 1);
					if ( !self.utils.isTransparent(backgroundColor) ) {
						var hexColor = self.utils.rgbToHex(backgroundColor),
							rgb = self.utils.hexToRgb(hexColor);

						if ( rgb ) {
							target.css('backgroundColor', 'rgba(' + rgb.r + ',' + rgb.g + ',' + rgb.b + ', ' + opac + ')');
						}
					}

					if ( target.find('> .is-background').length ) {
						target.find('> .is-background').css('opacity', opac);
					}

					// update data
					if ( cache.contents[type].hasOwnProperty(elID) && cache.contents[type][elID].hasOwnProperty('settings') ) {
						cache.contents[type][elID]['settings']['opacity']['opacity'] = opac;
					}
				}

				// fix background and corner
				if ( target.find('> .is-background').length && target.css('border') != 'none' ) {
					var styles = {};

					target.css('border', 'none');
					// border data
					if ( cache.contents[type][elID]['settings'].hasOwnProperty('border') ) {
						var borderSetting = cache.contents[type][elID]['settings']['border'];
						if ( borderSetting['mode'] == 'single' && borderSetting.hasOwnProperty('single') ) {
							for ( var key in borderSetting['single'] ) {
								styles[key] = borderSetting['single'][key];
							}
						} else {
							styles['border'] = borderSetting['all'];
						}
					}

					// corner data
					if ( cache.contents[type][elID]['settings'].hasOwnProperty('corner') ) {
						var cornerSetting = cache.contents[type][elID]['settings']['corner'];
						if ( cornerSetting['mode'] == 'single' && cornerSetting.hasOwnProperty('single') ) {
							for ( var key in cornerSetting['single'] ) {
								styles[key] = cornerSetting['single'][key];
							}
						} else {
							styles['borderRadius'] = cornerSetting['all'];
						}
					}
					target.find(" > .is-background").css(styles);
				}
			}
		});

		// update box element
		if ( self.frameBody.find('.ib2-box-el').length && !self.frameBody.find('.ib2-box-el').find('.ib2-box-content').length ) {
			self.frameBody.find('.ib2-box-el').each(function(){
				var box = $(this),
					elID = box.attr('id'),
					contentEl = self.frameBody.find('#' + elID + '-box'),
					height = contentEl.css('minHeight'),
					bgimg = '',
					bgvid = '',
					content = '';

				box.find('.ui-sortable').removeClass('ui-sortable');
				contentEl.parent().css('borderRadius', 0);
				bgimg = contentEl.append('<div class="ib2-box-content"></div>');
				if ( contentEl.find('> .ib2-element').length ) {
					contentEl.find('> .ib2-element').each(function(){
						var child = $(this);
						child.appendTo(contentEl.find('> .ib2-box-content'));
					});
				}

				contentEl.addClass('ib2-box-wrapper').css('height', 'auto');
				contentEl.parent().css('minHeight', height);
				contentEl.find('> .ib2-box-content').css('minHeight', height);
				contentEl.find('> .ib2-box-content').addClass('ib2-section-content');
				contentEl.removeClass('ib2-section-content');
			});
		}

		// fix notification bar
		if ( self.frameBody.find('.ib2-notification-text').length && cache.contents.hasOwnProperty('page') ) {
			var pageSettings = cache.contents['page']['screen-container']['settings'],
				barFont = "'Open Sans',sans-serif";

			if ( pageSettings && pageSettings.hasOwnProperty('pagetypo') && pageSettings['pageTypo'] ) {
				barFont = pageSettings['pageTypo']['fontFamily'] || "'Open Sans',sans-serif"; 
			}

			if ( pageSettings && pageSettings.hasOwnProperty('infobar') && pageSettings['infobar'] ) {
				if ( pageSettings['infobar'].hasOwnProperty('fontFamily') && pageSettings['infobar']['fontFamily'] && typeof pageSettings['infobar']['fontFamily'] === "string" ) {
					barFont = pageSettings['infobar']['fontFamily'] || "'Open Sans',sans-serif"; 
				}

				var barBgColor = pageSettings['infobar']['backgroundColor'] || '#5bc0de',
					barBorder = pageSettings['infobar']['border'] || '#FFFFFF',
					barColor = pageSettings['infobar']['color'] || '#FFFFFF';

				self.frameBody.find('.ib2-notification-text').removeAttr('style');
				self.frameBody.find('.ib2-notification-text').css({
					'fontFamily': barFont,
					'backgroundColor': barBgColor,
					'borderColor': barBorder,
					'color': barColor
				});
			}
		}

		// fix html code
		if ( self.frameBody.find('.ib2-code-el').length ) {
			self.frameBody.find('.ib2-code-el').each(function(){
				var el = $(this),
					target = el.find('> .el-content');

				target.css('min-height', '20px');
			});
		}

		// remove misplaced wide-box
		self.frameBody.find('.ib2-section-content').each(function(){
			if ( $(this).find('.ib2-wide-box').length ) {
				$(this).find('.ib2-wide-box').remove();
			}
		});

		var truHeight = 0;
		self.frameBody.find('#screen-container > .ib2-section-el').each(function(){
			var $this = $(this),
				height = $this.outerHeight();
			truHeight += height;
		});
		
		self.frameBody.find('.ib2-wide-box').each(function(){
			var $this = $(this),
				top = $this.css('top'),
				maxTop = truHeight + 50;

			if ( top ) {
				top = top.replace(/px/gi, '');
				top = parseInt(top);
				if ( top > maxTop ) {
					$this.remove();
				}
			}
		});

		// h-line cannot be wider than the parent
		self.frameBody.find('[data-el="hline"]').each(function(){
			var el = $(this),
				elID = el.attr('id'),
				parent = el.closest('.ib2-section-content');

			el.addClass('ib2-hline-el');
			if ( el.find('.ib2-hline').length && parent.length ) {
				var hwidth = el.find('.ib2-hline').outerWidth(),
					pwidth = parent.outerWidth(),
					pleft = parent.parent().css('paddingLeft'),
					pright = parent.parent().css('paddingRight');

				pleft = pleft ? pleft.replace(/px/gi, '') : 0;
				pright = pright ? pright.replace(/px/gi, '') : 0;

				var correctWidth = pwidth - pleft - pright;

				if ( hwidth >= correctWidth ) {
					el.find('.el-content').css('width', correctWidth + 'px');
					el.find('.ib2-hline').css('width', correctWidth + 'px');
				}
			}
		});

		// box cannot be wider than the page width
		self.frameBody.find('.ib2-box-el').each(function(){
			var el = $(this),
				elID = el.attr('id'),
				parent = el.closest('.ib2-section-content');

			if ( parent.length ) {
				var bwidth = el.find('> .el-content').outerWidth(),
					pwidth = self.pagewidth,
					pleft = parent.parent().css('paddingLeft'),
					pright = parent.parent().css('paddingRight');

				pleft = pleft ? pleft.replace(/px/gi, '') : 0;
				pright = pright ? pright.replace(/px/gi, '') : 0;

				var correctWidth = pwidth - pleft - pright;
				if ( bwidth >= correctWidth ) {
					el.find(' > .el-content').css('width', correctWidth + 'px');
					el.find(' > .el-content > .ib2-box-wrapper').css('width', correctWidth + 'px');
				}
			}
		});

		// disabled script from in html elements
		self.frameBody.find('.ib2-code-el > .el-content').find('script').each(function(){
			var script = $(this);
			if ( !script.attr('src') || script.attr('src') === '' ) {
				script.remove();
			}
		});

		// remove grammarly
		self.frameBody.find('grammarly-btn').remove();

		// fix background box data
		self.fixWideBox();

		// remove accidentally duplicated element item
		self.frameBody.find('#screen-container').find('.element-item').remove();

		// fill data in case there's a new element
		self.fillData();

		// update custom css with variables
		this.updateCssVariables();

		// new page settings
		self.newPageSettings();

		// remove powered by
		self.frameBody.find('#ib2-powered-by').remove();
	},

	newPageSettings: function() {
		var prime = this,
			cache = prime.cache,
			settings = cache.contents['page']['screen-container']['settings'];

		if ( !settings['pagetypo'].hasOwnProperty('fontFamily@active') ) {
			settings['pagetypo']['fontFamily@active'] = settings['pagetypo']['fontFamily'];
		}
	},

	fixWideBox: function(classOnly) {
		var prime = this,
			cache = prime.cache,
			settings = cache.contents,
			iframeDom = prime.frameBody;

		iframeDom.find('.ib2-wide-box').each(function(){
			var el = $(this),
				elID = el.attr('id');

			if ( classOnly ) {
				el.addClass('ib2-section-el');
			}

			var setData = false;
			if ( !settings.hasOwnProperty('wbox') ) {
				setData = true;
			}

			if ( settings.hasOwnProperty('wbox') && !settings.wbox.hasOwnProperty(elID) ) {
				setData = true;
			}

			if ( settings.hasOwnProperty('wbox') && settings.wbox.hasOwnProperty(elID) && settings.wbox[elID]['settings'] === null ) {
				setData = true;
			}

			if ( setData ) {
				prime.newElementData(elID, 'wbox');
			}
		});
	},

	renderElements: function() {
		var prime = this,
			gui = prime.gui,
			iframeDom = prime.frameBody,
			iframeHtml = prime.frameHtml,
			utils = prime.utils,
			cache = prime.cache;

		iframeDom.find('.ib2-comment-el').each(function(){
			if ( typeof cache.contents['comment'] === "undefined" ) {
				return true;
			}

			var el = $(this),
				elID = el.attr('id'),
				settings = cache.contents['comment'][elID]['settings'],
				provider = 'Facebook';

			if ( typeof settings['comment'] !== "undefined" && typeof settings['comment']['comment'] !== "undefined" ) {
				provider = utils.ucwords(settings['comment']['comment']);
			}

			if ( !el.find('> .el-content').length ) {
				el.html('<div class="el-content is-element element-placeholder" style="width:100%; max-width:100%; min-height: 30px; display:inline-block">{comment}</div>');
			}

			el.find('.el-content').html('<span class="area-hidden"><em>Comment Placeholder (selected system: <span class="comment-system content-preview">' + provider + '</span>)</em></span>');
		});

		iframeDom.find('.ib2-shortcode-el').each(function(){
			if ( typeof cache.contents['shortcode'] === "undefined" ) {
				return true;
			}

			var el = $(this),
				elID = el.attr('id'),
				settings = cache.contents['shortcode'][elID]['settings'],
				shortcode = '';

			if ( typeof settings['shortcode'] !== "undefined" && typeof settings['shortcode']['shortcode'] !== "undefined" ) {
				shortcode = $.trim(settings['shortcode']['shortcode']);
			}

			el.html('<span class="area-hidden"><em>Shortcode Placeholder (current shortcode: <span class="current-shortcode content-preview">' + shortcode + '</span>)</em></span>');
		});

		// fix bg color for bg image placeholder
		iframeDom.find('.is-background').each(function(i){
			$(this).css('backgroundColor', 'transparent');
		});
	},

	updateCssVariables: function() {
		var prime = this,
			gui = prime.gui,
			iframeDom = prime.frameBody,
			iframeHtml = prime.frameHtml,
			utils = prime.utils,
			cache = prime.cache;

		iframeDom.find('.ib2-button-el').each(function(){
			var el = $(this);
			if ( !el.hasClass('ib2-vars') ) {
				var elID = el.attr('id'),
					type = iframeDom.find('#' + elID).attr('data-el'),
					data = cache.contents[type][elID];

				if ( data.hasOwnProperty('settings') && data.settings ) {
					el.addClass('ib2-vars');
					var settings = data['settings'],	
						elem = iframeDom.find('#' + elID).find('.ib2-button'),
						style = settings['buttonstyle']['button-type'] || 'flat';

					iframeHtml.find('#' + elID + '-css').remove();
					elem.get(0).style.setProperty('--' + type + '-hover-color', settings['buttoncolor']['color@hover']);
					elem.get(0).style.setProperty('--' + type + '-hover-bg', settings['buttoncolor']['backgroundColor@hover']);
					elem.get(0).style.setProperty('--' + type + '-hover-tshadow', '1px 1px 0px ' + settings['buttoncolor']['textShadow@hover']);
					if ( style == 'glossy' ) {
						var colors = utils.getGradientColors(settings['buttoncolor']['backgroundColor@hover']);
						if ( colors ) {
							elem.get(0).style.setProperty('--' + type + '-hover-bgimg', 'linear-gradient(to bottom, ' + colors[0] + ', ' + colors[1] + ', ' + colors[2] + ')');
						} else {
							elem.get(0).style.setProperty('--' + type + '-hover-bgimg', 'none');
						}
					}

					if ( cache.cssRules.hasOwnProperty(elID) ) {
						delete cache.cssRules[elID];
					}
				}
			}
		});

		iframeDom.find('.ib2-form-el').each(function(){
			var el = $(this);
			if ( !el.hasClass('ib2-vars') ) {
				var elID = el.attr('id'),
					type = iframeDom.find('#' + elID).attr('data-el'),
					data = cache.contents[type][elID];

				if ( data.hasOwnProperty('settings') && data.settings ) {
					el.addClass('ib2-vars');
					var settings = data['settings'],
						setting = settings['formbutton'],
						elem = iframeDom.find('#' + elID).find('.ib2-button'),
						style = setting['submit-color-type'] || 'flat';

					iframeHtml.find('#' + elID + '-css').remove();
					elem.get(0).style.setProperty('--' + type + '-hover-color', setting['color@hover']);
					elem.get(0).style.setProperty('--' + type + '-hover-bg', setting['backgroundColor@hover']);
					elem.get(0).style.setProperty('--' + type + '-hover-tshadow', '1px 1px 0px ' + setting['textShadow@hover']);
					if ( style == 'glossy' ) {
						var colors = utils.getGradientColors(setting['backgroundColor@hover']);
						if ( colors ) {
							elem.get(0).style.setProperty('--' + type + '-hover-bgimg', 'linear-gradient(to bottom, ' + colors[0] + ', ' + colors[1] + ', ' + colors[2] + ')');
						} else {
							elem.get(0).style.setProperty('--' + type + '-hover-bgimg', 'none');
						}
					}

					if ( cache.cssRules.hasOwnProperty(elID) ) {
						delete cache.cssRules[elID];
					}
				}
			}
		});

		iframeDom.find('.ib2-menu-el').each(function(){
			var el = $(this);
			if ( !el.hasClass('ib2-vars') ) {
				var elID = el.attr('id'),
					type = iframeDom.find('#' + elID).attr('data-el'),
					data = cache.contents[type][elID];

				if ( data.hasOwnProperty('settings') && data.settings ) {
					el.addClass('ib2-vars');
					var settings = data['settings'],
						setting = settings['menustyle'],
						baseEl = el.find('.ib2-navi'),
						style = setting['style'] || 'plain';

					iframeHtml.find('#' + elID + '-css').remove();
					
					gui.setNavCss(style, baseEl, setting);

					if ( cache.cssRules.hasOwnProperty(elID) ) {
						delete cache.cssRules[elID];
					}
				}
			}
		});
	},

	fillData: function() {
		var self = this,
			iframeDom = this.frameBody,
			cache = this.cache;

		// base class for all elements
		iframeDom.find('.ib2-element').each(function(){
			var $el = $(this),
				elID = $el.attr('id'),
				type = $el.attr('data-el');

			if ( !cache.contents.hasOwnProperty(type) ) {
				cache.contents[type] = {};
			}

			if ( !cache.contents[type].hasOwnProperty(elID) ) {
				// need to save as new data
				self.newElementData(elID, type);
			}
		});

		// change optin to form element
		iframeDom.find('.ib2-optin-el').attr('data-el', 'form').addClass('ib2-form-el').removeClass('ib2-optin-el');
	},

	runCountdowns: function() {
		var prime = this,
			gui = prime.gui,
			iframeDom = prime.frameBody,
			cache = prime.cache,
			textSetting = cache.contents['page']['screen-container']['settings']['pagetypo'];

		iframeDom.find('.ib2-countdown-el').each(function(){
			var elID = $(this).attr('id');

			// update the content reveal style first
			if ( iframeDom.find('#' + elID + '-content').length ) {
				var paddingTop = iframeDom.find('#' + elID).css('paddingTop'),
					css = {
						paddingTop: paddingTop,
						fontSize: textSetting['fontSize'] + 'px',
							fontFace: textSetting['fontFace'],
							color: textSetting['color'],
							textShadow: 'none'
						};

				iframeDom.find('#' + elID + '-content').css(css).addClass('ib2-countdown-reveal');
			}

			var setting = cache.contents['countdown'][elID]['settings']['countdown'];
			if ( setting && setting.mode == 'date' ) {
				var offset = setting['tz'],
					hour = ("0" + setting['time-hour']).slice(-2),
					min = ("0" + setting['time-min']).slice(-2),
					newdate = setting['exact-date'] + ' ' + hour + ':' + min,
					origTime = moment.tz(newdate, offset),
					utc = origTime.clone().tz("UTC"),
					localTime = moment.utc(utc).toDate();

				finalDate = moment(localTime).format("YYYY/MM/DD HH:mm:ss");
			} else {
				var day = parseInt(setting['time-day']) * 24 * 60 * 60 * 1000,
					hour = parseInt(setting['time-hour']) * 60 * 60 * 1000,
					min = parseInt(setting['time-min']) * 60 * 1000,
					dateValue = day + hour + min;
				
				finalDate = new Date().valueOf() + dateValue;
			}

			// update countdown
			iframeDom.find('#' + elID).countdown(finalDate)
				.on('update.countdown', function(event) {
					time = {
						days: event.offset.totalDays, 
						hours: event.offset.hours, 
						minutes: event.offset.minutes, 
						seconds: event.offset.seconds
					};
					
					gui.countdownUpdate(iframeDom, elID, time, 'circular');
				});
		});
	},

	getCustomStyles: function() {
		var styles = window.FrameDocument.styleSheets,
			customStyles = {};

		for ( var i = 0; i < styles.length; i++ ) {
			if ( styles[i].href !== null ) continue;

			var rules = styles[i].cssRules,
				node = styles[i].ownerNode,
				elem = $(node),
				nodeId = elem.attr('id'),
				_nodeId = nodeId;

			if ( typeof nodeId === "undefined" ) {
				continue;
			}

			if ( !elem.hasClass('ib2-element-css') ) {
				continue;
			}

			nodeId = nodeId.replace('-css', '');
			nodeId = nodeId.replace('-submit', '');
			nodeId = nodeId.replace('-countdown', '');
			nodeId = nodeId.replace('-sub-hover', '');
			nodeId = nodeId.replace('-hover', '');
			
			if ( !customStyles.hasOwnProperty(nodeId) ) {
				customStyles[nodeId] = [];
			}

			for ( var j = 0; j < rules.length; j++ ) {
				// not css for an element?
				if ( rules[j].selectorText.indexOf('#' + nodeId) < 0 ) continue;

				// prepare data
				var newData = {}, style = rules[j].style,
					mode = rules[j].selectorText.indexOf('hover') > -1 ? 'hover' : 'normal',
					selectorText = rules[j].selectorText;

				newData['selector'] = rules[j].selectorText;
				newData['mode'] = mode;
				newData[mode] = {};
				for ( var key in style ) {
					if ( $.isNumeric(key) || style[key] == '' || typeof style[key] === "function" || typeof style[key] !== "string" ) continue;
					var cssKey = ( _nodeId.indexOf('sub') > -1 ) ? 'sub-' + key : key;
					if ( selectorText.indexOf('ib2-navi ul') > -1 || selectorText.indexOf('ib2-navi li ul') > -1 ) {
					 	cssKey = 'sub-' + key;
					}
					newData[mode][cssKey] = style[key];

				}
				customStyles[nodeId].push(newData);
			}
		}
		
		this.cache.cssRules = $.extend(true, {}, this.cache.cssRules, customStyles);
	},

	updateMiscElements: function() {
		var self = this,
			iframeDom = this.frameBody;

		// update shortcodes
		if ( iframeDom.find('#screen-container').find('.ib2-shortcode-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-shortcode-el').each(function(){
				var el = $(this),
					text = el.text(),
					id = el.attr('id');

				self.tempShortcode[id] = text;
				el.css('margin-bottom', '20px');
				el.addClass('element-placeholder')
					.html('<span class="area-hidden"><em>Shortcode Placeholder (current shortcode: <span class="current-shortcode content-preview">' + text + '</span>)</em></span>');
			});
		}

		// update html code elements
		if ( iframeDom.find('#screen-container').find('.ib2-code-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-code-el').each(function(){
				var el = $(this);
				el.find('> .el-content').css('height', 'auto');
			});
		}

		// update box elements
		if ( iframeDom.find('#screen-container').find('.ib2-box-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-box-el').each(function(){
				var el = $(this);
				el.find('> .el-content').css({
					'height': 'auto',
					'min-height': el.find('> .el-content').innerHeight() + "px"
				});

				el.find('> .el-content > .ib2-section-content').css({
					'height': 'auto',
					'min-height': el.find('> .el-content').innerHeight() + "px"
				});
			});
		}

		// update comments elements
		if ( iframeDom.find('#screen-container').find('.ib2-comment-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-comment-el').each(function(){
				var el = $(this),
					system = self.utils.ucwords(el.attr('data-comment'));

				el.css('margin-bottom', '20px');
				el.find('> .el-content')
					.addClass('element-placeholder')
					.html('<span class="area-hidden"><em>Comment Placeholder (selected system: <span class="comment-system content-preview">' + system + '</span>)</em></span>');
			});
		}

		// update share elements
		if ( iframeDom.find('#screen-container').find('.ib2-share-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-share-el').each(function(){
				var el = $(this),
					dir = $('#ib2-img-dir').val();

				el.css('margin-bottom', '20px');
				el.find('> .el-content').css({
					'display': 'flex',
					'width': '100%'
				});
				el.find('> .el-content > .ib2-facebook-share').html('<a href="#"><i class="fab fa-facebook-f"></i></a>');
				el.find('> .el-content > .ib2-twitter-share').html('<a href="#"><i class="fab fa-twitter"></i></a>');
				el.find('> .el-content > .ib2-linkedin-share').html('<a href="#"><i class="fab fa-linkedin-in"></i></a>');
				el.find('> .el-content > .ib2-google-share').html('<a href="#"><i class="fab fa-google-plus-g"></i></a>');

				el.find('.clearfix').remove();
			});
		}

		// update date element
		if ( iframeDom.find('#screen-container').find('.ib2-date-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-date-el').each(function(){
				var el = $(this);

				el.css('margin-bottom', '20px');
			});
		}

		// update menus element
		if ( iframeDom.find('#screen-container').find('.ib2-menu-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-menu-el').each(function(){
				var el = $(this),
					elID = el.attr('id'),
					style = el.attr('data-style'),
					menu = el.attr('data-menu'),
					html = $('#ib2-element-menu-html').html();

				menu = menu.replace(/'/g, '-039-');
				menu = menu.toLowerCase();
				menu = menu.replace(' ', '-');

				var _menu = menu;
				el.css('margin-bottom', '20px');
				if ( style && menu ) {
					menu = menu.toLowerCase();
					menu = menu.replace(" ", '-');
					var menuid = 'ib2-menu-' + menu + '-html',
						menuhtml = $('#' + menuid).length ? $('#' + menuid).html() : $('#ib2-default-menu-html').html();
				
					html = html.replace(new RegExp('{id}', 'gi'), elID);
					html = html.replace(new RegExp('{menu}', 'gi'), menuhtml);
					html = html.replace(new RegExp('ID-nav', 'gi'), elID + '-nav');

					el.replaceWith(html);

					iframeDom.find('#' + elID).attr('data-style', style);
					iframeDom.find('#' + elID).attr('data-menu', _menu);
					iframeDom.find('#' + elID).find('> nav > ul').removeClass('ib2-navi-plain ib2-navi-plain-pipe ib2-navi-flat-box ib2-navi-glossy');
					iframeDom.find('#' + elID).find('> nav > ul').addClass('ib2-navi-' + style);
					iframeDom.find('#' + elID).css('margin-bottom', '20px');
				}
				
			});
		}

		// update carousels
		this.updateCarousels();

		// update 3-step opt-ins
		this.update3Optins();

		// update tabs
		this.updateTabs();
	},

	updateQuestions: function() {
		var iframeDom = this.frameBody;

		// change quiz to questions element
		if ( iframeDom.find('#screen-container').find('.ib2-quiz-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-quiz-el').attr('data-el', 'questions').addClass('ib2-questions-el').removeClass('ib2-quiz-el');
		}

		if ( iframeDom.find('#screen-container').find('.ib2-questions-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-questions-el').each(function(){
				var $this = $(this),
					elID = $this.attr('id'),
					num = 0;

				$this.find('.ib2-quiz-page').each(function(i){
					if ( $(this).hasClass('ib2-section-content') ) {
						$(this).removeClass('ib2-unused-question');
					} else {
						if ( !$(this).hasClass('ib2-unused-question') ) {
							num++;
						}
						$(this).attr('data-question', (i+1));
					}
				});

				$this.attr('data-questions', num);
				$this.css('margin-bottom', '20px');
			});
		}

		// optin 3 element drop fix
		iframeDom.find('.ib2-section-slide').each(function(){
			var $this = $(this),
				slide = 1;

			$this.css('min-height', '0px');
			$this.css('height', '0px');

			if ( $this.is(":visible") ) {
				$this.css('height', 'auto');
				slide = parseInt($this.attr('data-slide'));
			}
			$this.parent().attr('data-current-slide', slide);
		});
	},

	updateTabs: function() {
		var iframeDom = this.frameBody;
		if ( iframeDom.find('#screen-container').find('.ib2-tabs-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-tabs-el').each(function(){
				var el = $(this);
				
				el.css('margin-bottom', '20px');

				el.find("> ul.nav-tabs > li.active").removeClass('active').addClass('ib2-tabs-active');
				el.find("> ul.nav-tabs").find('[data-toggle]').removeAttr('data-toggle');
				el.find("> ul.nav-tabs").removeClass('nav nav-tabs').addClass('ib2-nav-tabs');
				el.find("> .tab-content > .active").removeClass('active in').addClass('ib2-tabs-active');
				el.find("> .tab-content > .tab-pane").removeAttr('style');
				el.find("> .tab-content > .tab-pane").removeClass('tab-pane fade').addClass('ib2-tab-pane');
				el.find("> .tab-content").removeClass('tab-content').addClass('ib2-tab-content');
			});
		}
	},

	update3Optins: function() {
		var iframeDom = this.frameBody;
		if ( iframeDom.find('#screen-container').find('.ib2-optslide-el').length ) {
			iframeDom.find('#screen-container').find('.ib2-optslide-el').each(function(){
				var el = $(this);
				
				el.css('margin-bottom', '20px');
				el.find('> .ib2-section-slide').each(function(i){
					if ( i < 1 ) {
						$(this).show();
					} else {
						$(this).hide();
					}
				});
			});
		}
	},

	updateCarousels: function() {
		var defaultImage = $('#ib2-slides-src').attr('src'),
			iframeDom = this.frameBody;
		if ( iframeDom.find('#screen-container').find('.ib2-slide-container').length ) {
			iframeDom.find('#screen-container').find('.ib2-slide-container').each(function(){
				var el = $(this),
					elID = el.attr('id'),
					data = ib2globalsettings['carousel_data'][elID] || null,
					html = '';

				el.css('margin-bottom', '20px');
				if ( data ) {
					html += '<div id="' + elID + '-slides" class="carousel slide" data-ride="carousel">';
						html += '<ol class="carousel-indicators">';

						data.forEach(function(slide, i){
							var activeClass = ( i > 0 ? '' : ' class="active"' );
							html += '<li data-target="#' + elID + '-slides" data-slide-to="' + i + '"' + activeClass +'></li>';
						});
						html += '</ol>';

						html += '<div class="carousel-inner" role="listbox">';
						data.forEach(function(slide, i){
							if ( data[i]['imageurl'] == '' ) {
								ib2globalsettings['carousel_data'][elID][i]['imageurl'] = defaultImage;
							}
							var activeClass = ( i > 0 ? '' : ' active' ),
								img = data[i]['imageurl'],
								title = data[i]['title'],
								url = data[i]['desturl'];

							html += '<div class="item' + activeClass + '">';
								if ( url !== '' ) html += '<a href="' + url + '">';
	      						html += '<img src="' + img + '" alt="' + title + '">';
	      						if ( url !== '' ) html += '</a>';
	      						if ( title !== '' ) {
		      						html += '<div class="carousel-caption">' + title + '</div>';
	    						}
	    					html += '</div>';
						});
						html += '</div>';

						html += '<a class="left carousel-control" href="#' + elID + '-slides" role="button" data-slide="prev">';
    						html += '<span class="glyphicon glyphicon-chevron-left" aria-hidden="true"></span>';
    						html += '<span class="sr-only">Previous</span>';
  						html += '</a>';
  						html += '<a class="right carousel-control" href="#' + elID + '-slides" role="button" data-slide="next">';
    						html += '<span class="glyphicon glyphicon-chevron-right" aria-hidden="true"></span>';
    						html += '<span class="sr-only">Next</span>';
  						html += '</a>';

					html += '</div>';
				}

				el.find('> .el-content').html(html);
			});
		}
	},

	updateCountdowns: function() {
		var self = this,
			iframeDom = this.frameBody,
			gui = this.gui,
			ctype = 'normal',
			time = {};

		iframeDom.find('.ib2-countdown-el').each(function(){
			self.updateSingleCountdown($(this));
		});
	},

	updateSingleCountdown: function(el) {
		var iframeDom = this.frameBody,
			gui = this.gui,
			ctype = 'normal',
			time = {};

		var $c = el,
			elID = $c.attr('id'),
			mode = $c.data('mode'),
			target = $c.data('target'),
			style = $c.data('style'),
			hidden = '';
				
		// update content reveal
		if ( iframeDom.find('#' + elID + '-content') ) {
			hidden = iframeDom.find('#' + elID + '-content').html();
			iframeDom.find('#' + elID + '-content').remove();
		}

		$c.css('margin-bottom', '20px');
		if ( mode != 'date' ) {
			var parts = target.split(":"),
				day = parseInt(parts[0]),
				hour = parseInt(parts[1]),
				min = parseInt(parts[2]),
				newValue = day + hour + min;
			
			target = new Date().valueOf() + newValue;
		}
		
		if ( style == 'circular' ) {
			var format = this.getHtml('circularCountdown');
			format = format.replace(new RegExp('{id}', 'gi'), elID);
			ctype = style;
			iframeDom.find('#' + elID).html(format);

			// set stroke
			var radius = 54,
				circumference = 2 * Math.PI * radius;

			iframeDom.find('#' + elID).find('.clock-anim').css('stroke-dasharray', circumference);
			iframeDom.find('#' + elID).find('.clock-anim').css('stroke-dashoffset', circumference);
		} else {
			var format = this.getHtml('normalCountdown');
			format = format.replace(new RegExp('{id}', 'gi'), elID);
			iframeDom.find('#' + elID).html(format);

			if ( style == 'text' ) {
				if ( !iframeDom.find('#' + elID).find('.c-before-text').length ) {
					iframeDom.find('#' + elID).prepend('<span class="c-before-text"></span> ');
				}
				if ( iframeDom.find('#' + elID).attr('data-before') !== '' ) {
					iframeDom.find('#' + elID).find('.c-before-text').text(decodeURIComponent(iframeDom.find('#' + elID).attr('data-before')));
				}
				if ( !iframeDom.find('#' + elID).find('.c-after-text').length ) {
					iframeDom.find('#' + elID).append('<span class="c-after-text"></span> ');
				}
				if ( iframeDom.find('#' + elID).attr('data-after') !== '' ) {
					iframeDom.find('#' + elID).find('.c-after-text').text(decodeURIComponent(iframeDom.find('#' + elID).attr('data-after')));
				}
			}
		}

		// set margin
		iframeDom.find('#' + elID).css('margin', '10px auto 40px');

		// set hidden ?
		if ( hidden !== '' ) {
			iframeDom.find('#' + elID).append(hidden);
		}

		// update countdown
		iframeDom.find('#' + elID).countdown(target)
			.on('update.countdown', function(event) {
				time = {
					days: event.offset.totalDays, 
					hours: event.offset.hours, 
					minutes: event.offset.minutes, 
					seconds: event.offset.seconds
				};

				gui.countdownUpdate(iframeDom, elID, time, ctype);
			});
	},

	updateButtons: function() {
		var iframeDom = this.frameBody;
		iframeDom.find('.ib2-button-el').each(function(){
			var $this = $(this),
				elID = $this.attr('id'),
				btn = $this.find(' > .el-content > .ib2-button'),
				h = btn.innerHeight();

			btn.css('lineHeight', h + 'px');
			$this.css('margin-bottom', '14px');

			// fix width & height
			$this.find(' > .el-content').css('min-width', 0);
			$this.find(' > .el-content').css('width', btn.outerWidth() + 'px');
			$this.find(' > .el-content').css('min-height', $this.find(' > .el-content').height() + 'px');
			$this.find(' > .el-content').css('height', 'auto');

			btn.css('min-width', 0);
			btn.css('width', btn.outerWidth() + 'px');

			if ( !btn.find('.ib2-btn-txt').length ) {
				var label = btn.text(), 
					iconBefore = '', 
					iconAfter = '';

				if ( btn.find(' > i').length ) {
					iconBefore = btn.find(' > i').attr('class');
				}

				if ( btn.last().prop('tagName').toLowerCase() == 'i' ) {
					iconAfter = btn.last().attr('class');
				}

				btn.html('<span class="ib2-btn-txt">' + label + '</span>');
				if ( iconBefore !== '' ) {
					btn.prepend('<i class="' + iconBefore + '"></i>');
				}

				if ( iconAfter !== '' ) {
					btn.append('<i class="' + iconAfter + '"></i>');
				}
			}

			if ( btn.hasClass('open-popup') || btn.hasClass('ib2-open-popup') ) {
				btn.attr('data-popup', elID + '-popup');
			}
		});
	},

	updateBoxes: function() {
		var iframeDom = this.frameBody;
		if ( iframeDom.find('.ib2-box-el').length && !iframeDom.find('.ib2-box-el').find('.ib2-box-content').length ) {
			iframeDom.find('.ib2-box-el').each(function(){
				var box = $(this),
					elID = box.attr('id'),
					contentEl = iframeDom.find('#' + elID + '-box'),
					height = contentEl.css('minHeight'),
					bgimg = '',
					bgvid = '',
					content = '';

				box.find('.ui-sortable').removeClass('ui-sortable');
				contentEl.parent().css('borderRadius', 0);
				bgimg = contentEl.append('<div class="ib2-box-content"></div>');
				if ( contentEl.find('> .ib2-element').length ) {
					contentEl.find('> .ib2-element').each(function(){
						var child = $(this);
						child.appendTo(contentEl.find('> .ib2-box-content'));
					});
				}

				contentEl.addClass('ib2-box-wrapper');
				contentEl.parent().css('minHeight', height);
				contentEl.find('> .ib2-box-content').css('minHeight', height);
				contentEl.find('> .ib2-box-content').addClass('ib2-section-content');
				contentEl.removeClass('ib2-section-content');
			});
		}
	},

	updateImages: function() {
		var iframeDom = this.frameBody;
		iframeDom.find('.ib2-image-el').each(function(){
			var $this = $(this),
				img = $this.find('img'),
				src = img.attr('src'),
				w = img.width(),
				styles = ['img-rounded', 'img-circle', 'img-thumbnail'],
				styleclass = '',
				parentWidth = $this.parent().width();

			$this.css('margin-bottom', '20px');
			styles.forEach(function(sty) {
				if ( img.hasClass(sty) ) {
					styleclass = ' class="' + sty + '"';
				}
			});

			var caption = ( $this.find('.el-caption').length ) ? '<figcaption class="img-caption-txt">' + $this.find('.el-caption').text() + '</figcaption>' : '',
				figStyle = '';

			if ( $this.find('.el-caption').length && caption !== '' ) {
				figStyle += ' style="';
				figStyle += ' background-color: ' + $this.find('.el-caption').css('backgroundColor') + '; ';
				figStyle += ' color: ' + $this.find('.el-caption').css('color') + '; ';
				figStyle += '"';
			}

			var href = '',
				hrefClose = '';
			if ( $this.find('a').length ) {
				var a = $this.find('a');
				href += '<a href="' + a.attr('href') + '" target="' + ( a.attr('target') || '_self' ) + '">';
				hrefClose = '</a>';
			}

			var newImg = 
						'<div class="el-content is-element" style="width: ' + w + 'px; max-width: 100%">\
						<figure ' + figStyle + '>\
						' + href + '\
						<picture>\
							<source srcset="' + src + '" media="(min-width: 600px)">\
							<img src="' + src + '" style="width: ' + w + 'px; height: auto"' + styleclass + '>\
						</picture>\
						' + hrefClose + '\
						' + caption + '\
						</figure>\
						</div>';

			$this.html(newImg);
		});
	},

	updateVideos: function() {
		var self = this,
			iframeDom = this.frameBody,
			cache = this.cache;

		iframeDom.find('.ib2-video-container').each(function(){
			var elem = $(this),
				elID = elem.attr('id'),
				videos = ib2globalsettings.hasOwnProperty('videos') ? ib2globalsettings.videos : null,
				data = videos !== null && videos.hasOwnProperty(elID) ? videos[elID] : null,
				code = '';
			if ( data !== null ) {
				if ( data.hasOwnProperty('type') ) {
					if ( data.type == 'embed' && data.hasOwnProperty('embedcode') ) {
						try {
							code = window.atob(data.embedcode);
						} catch(e) {
							code = data.embedcode;
						}
	
						code = self.decodeHTML(code);
						elem.find(' > .el-content').html(code);
						elem.find(' > .el-content').find('iframe, embed').addClass('embed-responsive-item');
					} else {
						var codekey = (data.type == 'hosted') ? 'hostedcode' : data.type + 'code';
						var base_code = ( typeof data[codekey] === "string" && data[codekey] !== '' ) ? data[codekey] : '';
						try {
							code = window.atob(base_code);
						} catch(e) {
							code = base_code;
						}
						code += '&enablejsapi=1';
						code += ( data.type == 'youtube' ? '&mute=1' : '&muted=1' );
						code += '&otopelay=' + data.autoplay;
						code += ( data.type == 'youtube' && data.controls == 1 ? '&controls=1' : '' );
						code = code.replace('ib2player', 'videojs');
						code = code.replace(/^http:\/\//i, '//');
						code = code.replace('&autoplay=0', '');
						code = code.replace('&autoplay=1', '');

	 					var output = 
						'<div class="embed-responsive embed-responsive-16by9">\
							<iframe class="embed-responsive-item" src="' + code + '" scrolling="no" allowfullscreen webkitallowfullscreen mozallowfullscreen></iframe>\
						</div>';

						elem.find(' > .el-content').html(output);
					}
				}
			}
		});
	},

	updateSections: function() {
		var self = this,
			iframeDom = this.frameBody;

		iframeDom.find('.ib2-section-el').each(function(){
			var container = $(this),
				width = self.pagewidth,
				elID = container.attr('id');

			if ( container.hasClass('container') ) {
				container.removeClass('container');
			}

			// convert to columns element
			if ( container.parent().attr('id') != 'screen-container' ) {
				container.attr('data-el', 'columns').addClass('ib2-columns-el');
			}

			// Fix width
			if ( !container.hasClass('ib2-popup') ) {
				if ( container.attr('data-el') == 'section' ) {
					container.css('width', width + 'px');
					if ( container.find('> .el-content > .ib2-section-content').length == 1 ) {
						container.find('> .el-content > .ib2-section-content').css('width', width + 'px');
					}
				} else if ( container.attr('data-el') == 'columns' ) {
					container.css('width', '100%');
					if ( container.find('> .el-content > .ib2-section-content').length == 1 ) {
						container.find('> .el-content > .ib2-section-content').css('width', '100%');
					}
				} else {
					container.find('> .el-content > .el-content-inner').css('width', width + 'px');
					if ( container.find('> .el-content > .el-content-inner > .el-cols > .ib2-section-content').length == 1 ) {
						container.find('> .el-content > .el-content-inner > .el-cols > .ib2-section-content').css('width', width + 'px');
					}
				}
			}

			if ( elID == 'ib2_el_section1' && container.find('> .el-content > .ib2-section-content').length ) {
				var curHeight = container.find('> .el-content > .ib2-section-content').outerHeight();
				if ( curHeight > 1000 ) {
					var sectionHeight = curHeight - 400;
					container.find('> .el-content > .ib2-section-content').css('minHeight', sectionHeight + 'px');
				}
			}
		});

		// update the height as well
		iframeDom.find('.ib2-section-content').each(function(){
			$(this).css('min-height', $(this).outerHeight() + "px");
		});

		// unwrap & update bottom slider
		if ( iframeDom.find('#ib2-bottom-slider').length ) {
			var slider = iframeDom.find('#ib2-bottom-slider').html();

			iframeDom.find('#ib2-bottom-slider').remove();
			iframeDom.find('#screen-container').append(slider);
		}

		// if there's no bottom slider
		if ( !iframeDom.find('#ib2-bottom-slider-main').length ) {
			var newSlider = $('#ib2-slider-html').html();
			iframeDom.find('#screen-container').append(newSlider);
		}

		// fix bottom slider shadow
		iframeDom.find('#ib2-bottom-slider-main > .el-content').css('box-shadow', '0px 0px 6px 2px rgb(0,0,0,0.1)').addClass('is-element');
		
		// fix bottom slider width
		iframeDom.find('#ib2-bottom-slider-main-box').css('width', '100%');
		iframeDom.find('#ib2-bottom-slider-main').addClass('ib2-element');
		iframeDom.find('.container').removeClass('container');
		iframeDom.find('.el-content-inner').css('width', self.pagewidth + "px");

		// fix duplicate slider
		iframeDom.find('#screen-container').find('.ib2-slider-el').each(function(i){
			if ( i > 0 ) {
				$(this).remove();
			}
		});
	},

	updateElements: function() {
		var iframeDom = this.frameBody;
		// base class for all elements
		iframeDom.find('.ib2-section-el, .ib2-content-el').each(function(){
			var $el = $(this);

			$el.addClass('ib2-element');
		});

		// change optin to form element
		iframeDom.find('.ib2-optin-el').attr('data-el', 'form').addClass('ib2-form-el').removeClass('ib2-optin-el');
	},

	updateForSelection: function() {
		var iframeDom = this.frameBody;
		// for selections
		var oldElementClasses = '.el-content, .ib2-text-el, .ib2-title-el, .ib2-countdown-el, .ib2-menu-el, .ib2-quiz-el, .ib2-tabs-el, .ib2-optslide-el, .ib2-date-el, .ib2-shortcode-el';
		iframeDom.find(oldElementClasses).addClass('is-element');
	},

	updateForms: function() {
		var iframeDom = this.frameBody;
		iframeDom.find('.ib2-form-el').each(function(){
			var $el = $(this), 
				mode = $el.attr('data-form-mode') || 'vertical';

			$el.css('margin-bottom', '18px');

			// fix buttons class
			if ( !$el.find('form').find('.button-fields').length ) {
				var btns = $el.find('form').html();

				$el.find('form').find('button').remove();
				$el.find('form').find('input[type="image"]').remove();
				$el.find('form').append('<div class="button-fields"></div>');
				$el.find('form').find('.button-fields').html(btns);
				$el.find('form').find('.button-fields .form-fields').remove();
				$el.find('form').find('.button-fields .eu-checkbox').remove();
			}

			// add correct classes to button
			$el.find('.button-fields > button').attr('class', 'ib2-button ib2-form-submit');
			if ( !$el.find('.button-fields > button').find('.ib2-btn-txt').length ) {
				var btnTxt = $el.find('.button-fields > button').text();
				$el.find('.button-fields > button').html('<span class="ib2-btn-txt">' + btnTxt + '</span>');
			}

			// fix form classes
			$el.find('form').addClass('row').removeClass('form-inline');
			if ( mode == 'horizontal' ) {
				$el.find('form').find('.form-fields').attr('class', 'form-fields col-sm-12 col-md-8 inline-fields');
				$el.find('form').find('.button-fields').attr('class', 'button-fields col-sm-12 col-md-4 button-inline');
			} else if ( mode == 'semi-horizontal' ) {
				$el.find('form').find('.form-fields').attr('class', 'form-fields col-sm-12 col-md-12 inline-fields');
				$el.find('form').find('.button-fields').attr('class', 'button-fields col-sm-12 col-md-12 button-halfline');
			} else {
				$el.find('form').find('.form-fields').attr('class', 'form-fields col-sm-12 col-md-12');
				$el.find('form').find('.button-fields').attr('class', 'button-fields col-sm-12 col-md-12');
			}

			if ( $el.find('.ib2-form-image').length && $el.find('.ib2-form-image').attr('src') == '' ) {
				$el.find('.ib2-form-image').attr('src', $('#ib2-submit-img').attr('src'));
			}

			// fix fb style
			if ( $el.find('.ib2-facebook-optin-txt').attr('style') ) {
				var fbstyle = $el.find('.ib2-facebook-optin-txt').attr('style'),
					fbhidden = $el.find('.ib2-facebook-optin').is(":visible") ? false : true;
				$el.find('.ib2-facebook-optin-txt').removeAttr('style');
				$el.find('.ib2-facebook-optin').attr('style', fbstyle);

				if ( fbhidden ) {
					$el.find('.ib2-facebook-optin').hide();
				} else {
					$el.find('.ib2-facebook-optin').show();
				}
			}
		});
	},

	updateColumns: function() {
		var self = this,
			columns = window.FrameDocument.querySelectorAll(".el-cols > .ib2-section-content");

		for ( var i = 0; i < columns.length; i++ ) {
   			var col = columns[i];
   			var parent = col.parentNode;
   			var child = parent.firstChild;
   			var index = 0;

   			while (true) {
      			if ( child.nodeType === Node.ELEMENT_NODE ) index++;
      			if ( child === col || !child.nextSibling ) break;

      			child = child.nextSibling;
   			}

   			var styles = self.getStyles(col, ['width', 'marginRight']);

   			// get width for odd columns
		   	if ( index % 2 != 0 ) {
		      	var size = col.getBoundingClientRect();
		      	if ( 'width' in size ) {
		      		styles.width = ( Math.round((size.width * 100) / 100) ) + 'px';
		      	}
		   	}
		   	
		   	// clear the style first
		   	col.setAttribute('style', '');
		   	// update the column style
		   	for ( var prop in styles ) {
		   		col.style[prop] = styles[prop];
		   	}
		}
	},

	updateBackgrounds: function() {
		var self = this,
			iframeDom = this.frameBody;

		iframeDom.find('.ib2-element').each(function(){
			var $this = $(this),
				type = $this.attr('data-el'),
				elID = $this.attr('id'),
				target, 
				selector;

			if ( type == 'section' || type == 'wsection' || type == 'columns' || type == 'box' || type == 'wbox' ) {
				target = ( type == 'box' ) ? iframeDom.find('#' + elID + '-box') : iframeDom.find('#' + elID + ' > .el-content');
				selector = window.FrameDocument.getElementById(elID + '-box-img');
				oldBackgroundColor = target.css('backgroundColor');

				// check if for background placeholder if exists
				if ( iframeDom.find('#' + elID + '-box-img').length ) {
					// target must not have background image
					target.css('backgroundImage', 'none');

					iframeDom.find('#' + elID + '-box-img').addClass('is-background');
					if ( selector ) {
						var parentStyles = self.getStyles(selector.parentNode);
						parentStyles.width = '100%';
						parentStyles.maxWidth = '100%';
						if ( parentStyles ) {
							for ( var key in parentStyles ) {
								if ( key.indexOf('border') > -1 ) {
									iframeDom.find('#' + elID + '-box-img').css(key, parentStyles[key]);
								}
							}
						}
					}
				} else {
					// check if target has background
					var bgImg = target.css('backgroundImage');
					if ( bgImg !== 'none' ) {
						var bgPos = target.css('backgroundPosition'),
							bgAttach = target.css('backgroundAttachment'),
							bgTile = target.css('backgroundRepeat'),
							bgSize = target.css('backgroundSize'),
							opac = target.css('opacity'),
							imgUrl = /url\(\s*(['"]?)(.*?)\1\s*\)/g.exec(bgImg);

						imgUrl = imgUrl ? imgUrl[2] : '';

						// target must not have background image
						target.css('backgroundImage', 'none');

						// create new background image placeholder
						target.prepend('<div id="' + elID + '-box-img" class="is-background"></div>');

						var css = {
							backgroundColor: 'transparent',
							backgroundImage: 'url("' + imgUrl + '")',
							backgroundRepeat: bgTile,
							backgroundPos: bgPos,
							backgroundSize: bgSize,
							backgroundAttachment: bgAttach,
							position: 'absolute',
							top: 0,
							left: 0,
							maxWidth: '100%',
							width: '100%',
							height: target.css('height'),
							opacity: opac
						};

						if ( selector ) {
							var parentStyles = self.getStyles(selector.parentNode);
							if ( parentStyles ) {
								for ( var key in parentStyles ) {
									if ( key.indexOf('border') > -1 ) {
										css[key] = parentStyles[key];
									}
								}
							}
						}
						
						// apply background image
						iframeDom.find('#' + elID + '-box-img').css(css);
					}
				}
			}
		});
	},

	getStyles: function(el, excludes) {
		excludes = excludes || [];

		if ( el === null || el === false || typeof el === "undefined" ) {
			return null;
		}

		var styles = {},
   			elementStyle = ( el.style ? el.style : null );

   		if ( elementStyle ) {
   			for ( var prop in elementStyle ) {
   				if ( elementStyle.hasOwnProperty(prop) &&
   					typeof elementStyle[prop] !== "undefined" && 
   					elementStyle[prop] !== '' && 
   					elementStyle[prop] !== null &&
   					!$.isNumeric(prop) && 
   					$.inArray(prop, excludes) < 0 ) {
 
   					styles[prop] = elementStyle[prop];
   				}
   			}
   		}

   		return ( elementStyle ? styles : null );
	},

	getStyle: function(el, styleProp, setting) {
		var value = null, 
			isHover = false, 
			isCss = false,
			iframeDom = this.frameBody,
			cssRules = this.cache.cssRules;

		if ( !el.length ) {
			return value;
		}

		// special cases first
		var elID = ( el.hasClass('ib2-element') ? el.attr('id') : el.closest('.ib2-element').attr('id') ),
			type = iframeDom.find('#' + elID).attr('data-el'),
			boxed = ['section', 'wsection', 'box', 'columns', 'wbox'];

		if ( styleProp == 'opacity' && $.inArray(type, boxed) > -1 ) {
			if ( iframeDom.find('#' + elID + '-box-img').length ) {
				return iframeDom.find('#' + elID + '-box-img').css('opacity');
			} else if ( el.find(" > .ib2-bgvideo-param").length ) {
				return el.find(" > .ib2-bgvideo-param").css('opacity');
			} else {
				var bgColor = el.css('backgroundColor'),
					rgb = bgColor.match(/[.?\d]+/g);

				if ( rgb && typeof rgb[3] !== "undefined" ) {
					if ( rgb[0] == 0 && rgb[1] == 0 && rgb[2] == 0 && rgb[3] == 0 ) {
						return 1;
					} else {
						return rgb[3];
					}
				} else {
					return 1;
				}
			}
		}

		// background image
		if ( styleProp.indexOf('background') > -1 && styleProp !== 'backgroundColor' ) {
			if ( el.find(" .is-background").length ) {
				el = el.find(" .is-background");
			}
		}

		if ( styleProp == 'lineHeight' || (styleProp == 'paddingBottom' && setting.hasOwnProperty('props') && !setting.props.hasOwnProperty('single')) ) {
			el = iframeDom.find('#screen-container').find('p');
			if ( styleProp == 'lineHeight' ) {
				var _lh = el.css('lineHeight');
				if ( _lh && _lh.indexOf('px') > -1 ) {
					var _lh = parseFloat(_lh.replace(/px/gi, '')),
						_fs = parseInt(iframeDom.find('#screen-container').css('fontSize').replace(/px/gi, ''));

					_lh = (_lh/_fs);
					_lh = _lh.toFixed(1);
					return _lh;
				} else {
					return parseFloat(_lh);
				}
			}
		}

		// box shadow
		if ( styleProp.indexOf('boxShadow') > -1 ) {
			var shadow = el.css('boxShadow'),
				part = styleProp.replace('boxShadow', ''),
				realProp = ( styleProp === 'boxShadow' ? styleProp : part.toLowerCase() );

			return this.getBoxShadowStyle(shadow, realProp);
		}

		// background video
		if ( styleProp == 'backgroundVideo' ) {
			return ( el.find(" > .ib2-bgvideo-param").length ? 'yes' : 'no' );
		}

		// show delay effect
		if ( styleProp == 'showDelay' ) {
			return ( el.attr('data-delay') && el.attr('data-delay') != 'none' ? 'yes' : 'no' );
		}

		var _styleProp = styleProp;
		if ( styleProp.indexOf('@') > -1 ) {
			var styleParts = styleProp.split('@');
			if ( styleParts.length == 2 ) {
				styleProp = styleParts[0];
				if ( styleParts[1] == 'hover' ) {
					isHover = true;
					isCss = true;
				} else if ( styleParts[1] == 'active' ) {
					isHover = false;
					isCss = true;
				}
			}
		}

		if ( !isCss && setting.hasOwnProperty('customCSS') && $.inArray(styleProp, setting['customCss']) ) {
			isCss = true;
		}

		value = el.css(styleProp);
		if ( type == 'tabs' && isCss && $.inArray(_styleProp, ['backgroundColor@active', 'backgroundColor']) > -1 && !cssRules.hasOwnProperty(elID) ) {
			value = '#ffffff';
		}

		if ( type == 'tabs' && isCss && _styleProp == 'backgroundColor@hover' && !cssRules.hasOwnProperty(elID) ) {
			value = '#e6e6e6';
		}

		if ( type == 'tabs' && isCss && _styleProp == 'color' && !cssRules.hasOwnProperty(elID) ) {
			value = '#555555';
		}

		if ( isCss && !cssRules.hasOwnProperty(elID) ) {
			// this maybe be a css variable
			var field;
			setting.fields.forEach(function(fld){
				if ( fld.hasOwnProperty('prop') && fld.prop == _styleProp ) {
					field = fld;
				}
			});

			if ( field && field.action == 'cssvar' ) {
				var varname = this.cssVarMap(type, field, _styleProp);
				if ( varname !== '' ) {
					var target = ( field.cssSelector == 'self' ) ? el.get(0) : el.find(' ' + field.cssSelector).get(0);
					if ( target ) {
						value = getComputedStyle(target).getPropertyValue(varname);
						value = $.trim(value);
					}
				}
			}
		}

		if ( isCss && cssRules.hasOwnProperty(elID) ) {
			// get css selector
			var defSelector = null;
			setting.fields.forEach(function(defField) {
				if ( defField.hasOwnProperty('prop') && defField.prop == _styleProp ) {
					defSelector = defField.cssSelector || null;
				}
			});

			if ( defSelector ) {
				defSelector = defSelector.replace(new RegExp('{elID}', 'g'), '#' + elID);
				for ( var c = 0; c < cssRules[elID].length; c++ ) {
					var realSelector = cssRules[elID][c]['selector'];
					if ( realSelector.indexOf(",") > -1 ) {
						var parts = realSelector.split(",");
						realSelector = $.trim(parts[0]);
					}
					//console.log('def:', defSelector, 'css:', realSelector);
					if ( defSelector == realSelector ) {
						//console.log('masup', cssRules[elID][c]);
						if ( isHover ) {
							if ( !cssRules[elID][c].hasOwnProperty('hover') ) {
								continue;
							}
							
							if ( cssRules[elID][c]['hover'].hasOwnProperty(styleProp) ) {
								value = cssRules[elID][c]['hover'][styleProp];
								break;
							}
						} else {
							if ( !cssRules[elID][c].hasOwnProperty('normal') ) {
								continue;
							}
							
							if ( cssRules[elID][c]['normal'].hasOwnProperty(styleProp) ) {
								value = cssRules[elID][c]['normal'][styleProp];
								break;
							} else {
								if ( styleProp == 'border' ) {
									if ( cssRules[elID][c]['normal'].hasOwnProperty('borderTop') ) {
										value = cssRules[elID][c]['normal']['borderTop'];
										break;
									} else if ( cssRules[elID][c]['normal'].hasOwnProperty('borderLeft') ) {
										value = cssRules[elID][c]['normal']['borderLeft'];
										break;
									}
								}
							}
						}
					}
				}
			}
		}

		// text shadow
		if ( $.inArray(styleProp, ['textShadowToggle', 'textShadowColor', 'textShadowBlur']) > -1 ) {
			var shadowcss = el.css('textShadow'),
				textShadow = shadowcss.match(/(-?\d+px)|(rgb\(.+\))|([a-zA-Z]+)/g);

			if ( styleProp == 'textShadowToggle' ) {
				return ( textShadow == 'none' ? 'no' : 'yes' );
			} else if ( styleProp == 'textShadowColor' ) {
				value = textShadow != 'none' && textShadow.length ? textShadow[0] : '#808080';
			} else {
				return 0;
			}
		}

		// parse text shadow
		if ( styleProp == 'textShadow' ) {
			var origValue = value,
				textShadow = value.match(/(-?\d+px)|(rgb\(.+\))|([a-zA-Z]+)/g);

			value = (textShadow && textShadow != 'none' ? textShadow[0] : '#808080');
			if ( value && value.indexOf('rgb') < 0 ) {
				var shadowParts = origValue.split(" ");
				value = (shadowParts && typeof shadowParts[3] !== "undefined" && shadowParts[3].indexOf('#') > -1) ? shadowParts[3] : '#808080';
			}
		}

		if ( typeof value !== "undefined" && value !== null && value !== '' ) {
			if ( styleProp == 'backgroundImage' ) {
				value = /url\(\s*(['"]?)(.*?)\1\s*\)/g.exec(value);
				value = value ? value[2] : '';
			}

			// get correct background color
			if ( styleProp == 'backgroundColor' && iframeDom.find('#' + elID + '-box-img').length ) {
				value = iframeDom.find('#' + elID + '-box-img').css('backgroundColor');
			}

			var rgbRegex = /^rgb\(\s*\d{1,3}\s*,\s*\d{1,3}\s*,\s*\d{1,3}\s*\)$/,
				rgbaRegex = /^rgba?[\s+]?\([\s+]?(\d+)[\s+]?,[\s+]?(\d+)[\s+]?,[\s+]?(\d+)[\s+]?/i;

			if ( rgbRegex.test(value) || rgbaRegex.test(value) ) {
				value = this.utils.rgbToHex(value);
			}

			var leaveUnit = ['border', 'borderLeft', 'borderTop', 'borderRight', 'borderBottom'];
			if ( typeof value === "string" && $.inArray(styleProp, leaveUnit) < 0 ) {
				value = value.replace(/px/gi, '');
			}

			if ( $.isNumeric(value) && setting.hasOwnProperty('fields') ) {
				var propField;
				setting.fields.forEach(function(field) {
					if ( styleProp == field.prop ) {
						propField = field;
					}
				});

				if ( propField && propField.hasOwnProperty('unit') && propField.unit == '%' && styleProp == 'width' ) {
					var parentWidth = el.parent().outerWidth();
					value = Math.round(value/parentWidth*100);
				}
				value = parseFloat(value);
			}

			if ( typeof value === "string" ) {
				value = value.replace(new RegExp('{%IB2_PLUGIN_DIR%}', 'g'), $('#ib2-plugin-url').val());
			}

			if ( styleProp == 'fontFamily' ) {
				value = value.replace(new RegExp('"', 'gi'), "'");
				value = value.replace(new RegExp(', ', 'g'), ',');
			}
		}
		return value;
	},

	getVideoData: function(el, fileid, setting) {
		var value = '',
			elID = ( el.hasClass('ib2-element') ? el.attr('id') : el.closest('.ib2-element').attr('id') );
		
		if ( ib2globalsettings.hasOwnProperty('videos') && ib2globalsettings['videos'].hasOwnProperty(elID) ) {
			var data = ib2globalsettings['videos'][elID],
				fileKey = fileid.replace('video-', ''),
				value = ( data.hasOwnProperty(fileKey) ? data[fileKey] : null );

			if ( fileid.indexOf('hosted') > -1 ) {
				fileKey = 'hosted';
			}

			switch ( fileKey ) {
				case 'youtube':
				case 'vimeo':
					value = data.url;
					break;

				case 'hosted':
					var subkey = fileid.replace('video-hosted-', '');
					value = data[subkey];
					break;

				case 'embed':
					if ( data.hasOwnProperty('embedcode') ) {
						try {
							value = window.atob(data.embedcode);
						} catch(e) {
							value = $.trim(data.embedcode);
						}
						value = this.decodeHTML(value);
					}
					break;
			}
		} else {
			// new element
			var fileKey = fileid.replace('video-', '');
			if ( fileKey == 'type' ) {
				value = 'youtube';
			} else if ( fileKey == 'autoplay' ) {
				value = 0;
			} else if ( fileKey == 'controls' ) {
				value = 1;
			} else {
				value = '';
			}
		}
		return value;
	},

	getFile: function(el, fileid, setting) {
		var value = null,
			iframeDom = this.frameBody,
			isBgVids = false,
			vids = ['mp4', 'ogg', 'webm'],
			delays = ['delayhour', 'delaymin', 'delaysec'],
			elID = ( el.hasClass('ib2-element') ? el.attr('id') : el.closest('.ib2-element').attr('id') );

		if ( !el.length ) {
			return value;
		}

		if ( fileid.indexOf('share-') > -1 ) {
			var social = fileid.replace('share-', '');
			if ( el.find('.ib2-' + social + '-share').length ) {
				return 'yes';
			} else {
				return 'no';
			}
		}

		if ( fileid == 'gdpr' ) {
			return (el && el.length ? 'yes' : 'no');
		}

		if ( fileid == 'shortcode' ) {
			if ( this.tempShortcode.hasOwnProperty(elID) ) {
				return this.tempShortcode[elID];
			}
		}

		if ( fileid == 'code' ) {
			return el.find('> .el-content').html() ? $.trim(el.find('> .el-content').html()) : '';
		}

		// FB Optin
		if ( fileid == 'facebook-enable' ) {
			return ( el.is(":visible") ? "yes" : "no" );
		}

		if ( fileid == 'facebook-label' ) {
			return el.find('.ib2-facebook-subscribe').text();
		}

		if ( fileid == 'facebook-text' ) {
			return el.find('.ib2-facebook-optin-txt').text();
		}

		if ( fileid == 'facebook-only' ) {
			return ( el.parent().find('.form-fields').is(":visible") ? "no" : "yes" );
		}

		var elID = ( el.hasClass('ib2-element') ? el.attr('id') : el.closest('.ib2-element').attr('id') );
		if ( fileid == 'button-type' || fileid == 'video-type' || fileid == 'form-mode' || fileid == 'style' ) {
			el = iframeDom.find('#' + elID);
		}

		if ( fileid == 'submit-color-type' ) {
			fileid = 'button-type';
		}

		if ( fileid == 'popup' && value === null ) {
			value = 'none';
		}

		if ( setting.id == 'countdown-settings' ) {
			var targetDate = el.attr('data-target'),
				parts = ( el.attr('data-mode') == 'date' ? targetDate.split(' ') : targetDate.split(':') );

			if ( el.attr('data-mode') == 'date' ) {
				var clocks = parts[1].split(":");
				if ( fileid == 'exact-date' ) {
					return parts[0].replace(new RegExp('/', 'g'), '-');
				} else if ( fileid == 'time-day' ) {
					return 0;
				} else if ( fileid == 'time-hour' ) {
					return parseInt(clocks[0]);
				} else if ( fileid == 'time-min' ) {
					return parseInt(clocks[1]);
				}
			} else {
				if ( fileid == 'exact-date' ) {
					return '2020-01-01';
				} else if ( fileid == 'time-day' ) {
					return (parseInt(parts[0]) / 1000 / 60 / 60 / 24);
				} else if ( fileid == 'time-hour' ) {
					return (parseInt(parts[1]) / 1000 / 60 / 60);
				} else if ( fileid == 'time-min' ) {
					return (parseInt(parts[2]) / 1000 / 60);
				}
			}
		}

		if ( fileid.indexOf('video') > -1 ) {
			return this.getVideoData(el, fileid, setting);
		}

		if ( $.inArray(fileid, vids) > -1 ) {
			isBgVids = true;
		}

		if ( fileid == 'animation' ) {
			value = 'none';
		}

		if ( $.inArray(fileid, delays) > -1 ) {
			if ( el.attr('data-delay') && el.attr('data-delay') != 'none' ) {
				var nums = { delayhour: 0, delaymin: 1, delaysec: 2 },
					parts = el.attr('data-delay').split(":");

				if ( parts.length ) {
					return parseInt(parts[nums[fileid]]);
				}
			}
			return 0;
		}

		if ( !isBgVids && el.attr('data-' + fileid) ) {
			value = el.attr('data-' + fileid);
		} else if ( isBgVids && el.find(" > .ib2-bgvideo-param").attr('data-' + fileid) ) {
			value = el.find(" > .ib2-bgvideo-param").attr('data-' + fileid);
		}

		if ( fileid == 'form-mode' && value === null ) {
			value = 'vertical';
		}

		if ( fileid == 'button-action' ) {
			if ( el.hasClass('open-popup') || el.hasClass('ib2-open-popup') ) {
				return 'popup';
			}
			if ( !value || value === null ) {
				return el.attr('href') && el.attr('href') != '' ? 'url' : 'none';
			}
		}
		return value;
	},

	getText: function(el, textClass, setting) {
		if ( textClass && el.find('.' + textClass).length ) {
			return el.find('.' + textClass).text();
		} else {
			if ( textClass == 'c-before-text' || textClass == 'c-after-text' ) {
				return '';
			}
			return el.text();
		}
	},

	getAttrib: function(el, attrib, setting) {
		if ( el.hasClass('ib2-form-submit') ) {
			el = el.parent().find('.ib2-form-image');
		}

		if ( attrib.indexOf('_') > -1 ) {
			if ( el.find('input[name="' + attrib + '"]').length ) {
				return el.find('input[name="' + attrib + '"]').val();
			} else {
				return ( attrib == '_webinar_key' ? 'none' : '' );
			}
		}

		var value = el.attr(attrib);
		if ( typeof value === "undefined" ) {
			value = '';
		}

		if ( value == '_blank' ) {
			value = 'yes';
		} else if ( value == '_self' ) {
			value = 'no';
		}
		
		return value;
	},

	getClass: function(el, classkey, setting) {
		var currentclass = 'none',
			_classkey = classkey;

		classkey = classkey + 'es';
		if ( setting.hasOwnProperty(classkey) ) {
			setting[classkey].forEach(function(classname) {
				if ( el.hasClass(classname) ) {
					currentclass = classname;
				}
			});
		} else {
			var tagname = el.prop("tagName") || 'DIV';
			tagname = tagname.toLowerCase();

			if ( tagname == 'i' ) {
				return el.attr('class');
			} else {
				// this is a checkbox
				return el.hasClass('.' + _classkey) ? "yes" : 'no';
			}
		}

		if ( el && el.prop('tagName').toLowerCase() == 'input' && currentclass == 'none' ) {
			if ( classkey == 'fieldstyleclasses' ) {
				currentclass = 'field-normal';
			} else if ( classkey == 'fieldclasses' ) {
				currentclass = 'input-md';
			}
		}
		return currentclass;
	},
	
	getBoxShadowStyle: function(boxShadow, shadowProp) {
		var defaults = {
			boxShadow: 'none',
			h: 0,
			v: 0,
			blur: 4,
			spread: 2,
			color: '#000000',
			alpha: 0.1
		};
		if ( boxShadow == 'none' ) {
			return defaults[shadowProp];
		} else {
			var shadow = boxShadow.match(/(-?\d+px)|(rgb\(.+\))|([a-zA-Z]+)/g),
				value = '';
			switch ( shadowProp ) {
				case 'alpha':
				case 'color':
					var rgbRegex = /^rgb\(\s*\d{1,3}\s*,\s*\d{1,3}\s*,\s*\d{1,3}\s*\)$/,
					rgbaRegex = /^rgba?[\s+]?\([\s+]?(\d+)[\s+]?,[\s+]?(\d+)[\s+]?,[\s+]?(\d+)[\s+]?/i;

					if ( rgbRegex.test(shadow[0]) || rgbaRegex.test(shadow[0]) ) {
						if ( shadowProp == 'alpha' ) {
							var parts = shadow[0].substring(shadow[0].indexOf("(")).split(",");
							value = parts[3] || 0.1;
						} else {
							value = this.utils.rgbToHex(shadow[0]);
						}
					} else {
						value = shadow[0];
					}
					break;
				case 'h':
					value = shadow[1].replace(/px/gi, '');
					break;
				case 'v':
					value = shadow[2].replace(/px/gi, '');
					break;
				case 'blur':
					value = shadow[3].replace(/px/gi, '');
					break;
				case 'spread':
					value = shadow[4].replace(/px/gi, '');
					break;
			}
			return value;
		}
	},

	getElementData: function(type, alias) {
		var data,
			key = ( alias ? 'alias' : 'type' ),
			val = ( alias ? alias : type ),
			elements = this.elements,
			cache = this.cache;

		// try to load from cache first
		if ( cache.elementData.hasOwnProperty(val) ) {
			return cache.elementData[val];
		}

		for ( var i = 0; i < elements.length; i++ ) {
			if ( elements[i].hasOwnProperty(key) && elements[i][key] == val ) {
				
				if ( key == 'type' && elements[i].hasOwnProperty('alias') ) {
					continue;
				}

				data = elements[i];
				cache.elementData[val] = data;
				break;
			}
		}
		return data;
	},

	applySettings: function(options) {
		var cache = this.cache,
			iframeDom = this.frameBody,
			defSettings = this.settings,
			elID = cache.activeElement,
			type = cache.activeElementType,
			data, settings;

		if (! cache.contents.hasOwnProperty(type) ) {
			return;
		}

		if (! cache.contents[type].hasOwnProperty(elID) ) {
			return;
		}

		data = cache.contents[type][elID];
		if ( data.hasOwnProperty('settings') ) {
			settings = data.settings;

			// check settings against options
			options.forEach(function(option){
				if ( !settings.hasOwnProperty(option.name) ) {
					updated = true;
					cache.contents[type][elID]['settings'][option.name] = { pos: option.target };
				}
			});

			for ( var key in settings ) {
				if ( defSettings.hasOwnProperty(key) && defSettings[key].hasOwnProperty('fields') ) {
					var def = defSettings[key],
						set = settings[key],
						fields = defSettings[key]['fields'];

					if ( $('#' + type + '-' + def.id).find('.settings-mode').length && set.hasOwnProperty('mode') && $.inArray(set.mode, ['single', 'all']) > -1) {
						this.toggleModeFields(type, def.id, set.mode);
					}

					// set values
					fields.forEach(function(field){
						if ( field['id'] == 'esp-integration' ) {
							if ( set.hasOwnProperty('esp-type') && set['esp-type'] !== 'none' ) {
								this.gui.updateListOptions(this.esplists[set['esp-type']]);
								$('.esp-refresh-btn').removeClass('hidden');
							} else {
								$('.esp-refresh-btn').addClass('hidden');
							}
						}

						if ( field.type == 'dynamic' ) {
							var fieldsID = 'fields-' + key + '-' + defSettings[key]['id'],
								values = set[key],
								total = values.length,
								html = $('#' + fieldsID).html();

							if ( html !== '' ) {
								$('#' + fieldsID).html('');
								for ( var i = 0; i < total; i++ ) {
									var output = html,
										num = i+1;

									if ( i > 0 ) {
										output = '<hr>' + output;
									}

									output = output.replace(new RegExp("{n}", 'gi'), num);
									$('#' + fieldsID).append(output);

									field.dynamic.forEach(function(input) {
										var fieldId = input.type + '-' + field.id + '-' + input['dynamic'] + '-' + num + '-' + type,
											val = values[i][input.dynamic];

										this.applyValue(fieldId, input.type, val);
										this.applySpecificSetting(fieldId, input.type, val);

										$('#' + fieldId).attr('data-num', i);
										$('#' + fieldId).attr('data-key', input.dynamic);
									}, this);
									
								}
							}
						} else {
							var settingData = set,
							fieldId = field.type + '-' + field.id + '-' + type;
							if ( set.hasOwnProperty('single') ) {
								settingData = set.single;
							}

							if ( field.hasOwnProperty('mode') && field.mode == 'all' ) {
								this.applyValue(fieldId, field.type, set.all);
								this.applySpecificSetting(fieldId, field.type, set.all);
							} else {
								var dataKey = field.hasOwnProperty('file') ? field.file : ( field.hasOwnProperty('textClass') ? field.textClass : ( field.attrib || field.prop ) ),
									defaultVal = ( field.hasOwnProperty('defaultValue') ? field.defaultValue : '' ),
									val = ( settingData.hasOwnProperty(dataKey) ) ? settingData[dataKey] : defaultVal;

								if ( typeof dataKey === "undefined" && field.hasOwnProperty('classname') ) {
									dataKey = field.classname;
									val = settingData[dataKey];
								}
								if ( typeof dataKey === "undefined" && field.hasOwnProperty('typename') ) {
									dataKey = field.typename;
									val = settingData[dataKey];
								}

								if ( val === null || val === '' || val == undefined ) {
									var funcName = this.settingFunctionMap(field),
										targetName = ( set.hasOwnProperty('pos') ? set.pos: 'outer' ),
										targetElem = this.getSettingTarget(elID, targetName);
										
									if ( funcName && ( !set.hasOwnProperty(dataKey) || set[dataKey] === null ) ) {
										val = this.readSetting(targetElem, dataKey, funcName, defSettings[key]);
										val = ( typeof val !== "undefined" && val !== null ) ? val : defaultVal;
										cache.contents[type][elID]['settings'][key][dataKey] = ( typeof val !== "undefined" ) ? val : defaultVal;
									} else {
										val = ( $.inArray(field.type, ['select', 'posts']) > -1 ) ? 'none' : defaultVal;
									}
								} else {
									// we need to check again if the data really exist
									if ( !cache.contents[type][elID]['settings'][key].hasOwnProperty(dataKey) ) {
										cache.contents[type][elID]['settings'][key][dataKey] = val;
									}
								}

								if ( val == 'inherit' ) {
									if ( dataKey == 'fontFamily' ) {
										if ( iframeDom.find('#' + elID).hasClass('ib2-title-el') ) {
											val = cache.contents['page']['screen-container']['settings']['pagetypo']['fontFamily@active'];
										} else {
											val = cache.contents['page']['screen-container']['settings']['pagetypo']['fontFamily'];
										}
									}
								}

								if ( field.type == 'checkbox' && val === 'false' ) {
									val = 'no';
								}

								if ( field.type == 'fontstyle' ) {
									val = [settingData['fontWeight'], settingData['fontStyle'], settingData['textDecorationLine']];
								}

								if ( field.hasOwnProperty('isURL') && field.isURL === true && typeof val === "string" && val !== '' && val !== null ) {
									if ( val.indexOf('#') > -1 && val.length == 1 ) {
										val = '';	
									} else {
										try {
											val = window.atob(val);
										} catch(e) {
											cache.contents[type][elID]['settings'][key][dataKey] = Base64.encode(val);
										}
									}
								}

								if ( field.hasOwnProperty('prop') && field.prop == 'fontFamily' ) {
									val = val.replace(new RegExp('"', 'gi'), "'");
									val = val.replace(", ", ",");
								}

								if ( dataKey && dataKey.indexOf('fontFamily') > -1 ) {
									val = val.replace(new RegExp(', ', 'g'), ',');
								}
								this.applyValue(fieldId, field.type, val);
								this.applySpecificSetting(fieldId, field.type, val);

								if ( field.type == 'posts' && field.hasOwnProperty('onChange') ) {
									if ( val !== '' && val != 'none' && $.isNumeric(val) && typeof field['onChange'] === "function" ) {
										field['onChange'](val);
									}
								}
							}
						}
					}, this);
				}
			}
		}

		if ( type == 'page' ) {
			if ( $('#ib2-gate-url').val() !== '' ) {
				$('#gateway-gateway-url-page').val($('#ib2-gate-url').val());
			}
		}
	},

	applyValue: function(id, type, val) {
		if ( typeof val === "string" ) {
			val = val.replace(new RegExp('{%IB2_PLUGIN_DIR%}', 'g'), $('#ib2-plugin-url').val());
		}

		if ( type == 'border' && typeof val === "string" && val !== '' ) {
			val = val.replace(new RegExp(', ', 'g'), ",");
			var vals = val.split(" ");

			if ( vals.length == 3 ) {
				var bWidth = vals[0].replace(/px/gi, ''),
					bStyle = vals[1],
					bColor = this.utils.rgbToHex(vals[2]);

				$('#' + id + '-width').val(bWidth);
				$('#' + id + '-width-num').val(bWidth);
				$('#' + id + '-style').val(bStyle);
				$('#' + id + '-color').val(bColor);

				var inputColor = this.utils.inputTextColor(bColor);
				$('#' + id + '-color').css({
					backgroundColor: bColor,
					color: inputColor
				});
			}
		} else if ( type == 'checkbox' ) {
			var reveal = $('#' + id).attr('data-reveal'),
				negative = ( $('#' + id).attr('data-negative') == 'yes' ? true : false );

			val = $.isNumeric(val) ? parseInt(val) : val;
			if ( val == 'yes' || val == 1 ) {
				if ( negative ) {
					$('#' + id).removeAttr('checked');
				} else {
					$('#' + id).attr('checked', 'checked');
				}
				if ( typeof reveal === "string" && reveal !== 'none' ) {
					$('#' + id).closest('.settings-fields').find('.' + reveal).removeClass('chk-hidden');
				}
			} else {
				if ( negative ) {
					$('#' + id).attr('checked', 'checked');
				} else {
					$('#' + id).removeAttr('checked');
				}
				if ( typeof reveal === "string" && reveal !== 'none' ) {
					$('#' + id).closest('.settings-fields').find('.' + reveal).addClass('chk-hidden');
				}
			}
		} else if ( type == 'fontstyle' && $.isArray(val) ) {
			if ( val[0] <= 400 || val[0] == '400' ) {
				val[0] = 'normal';
			}

			if ( val[0] != 'normal' ) {
				$('#' + id + '-bold').parent().addClass('active');
				$('#' + id + '-bold').attr('checked', 'checked');
			} else {
				$('#' + id + '-bold').parent().removeClass('active');
				$('#' + id + '-bold').removeAttr('checked');
			}

			if ( val[1] != 'normal' ) {
				$('#' + id + '-italic').parent().addClass('active');
				$('#' + id + '-italic').attr('checked', 'checked');
			} else {
				$('#' + id + '-italic').parent().removeClass('active');
				$('#' + id + '-italic').removeAttr('checked');
			}

			if ( val[2] != 'none' ) {
				$('#' + id + '-underline').parent().addClass('active');
				$('#' + id + '-underline').attr('checked', 'checked');
			} else {
				$('#' + id + '-underline').parent().removeClass('active');
				$('#' + id + '-underline').removeAttr('checked');
			}
		} else {
			$('#' + id).val(val);
		}
	},

	applySpecificSetting: function(id, type, val) {
		var elem = $('#' + id),
			tagName = elem.prop('tagName');

		if ( typeof val === "string" ) {
			val = val.replace(new RegExp('{%IB2_PLUGIN_DIR%}', 'g'), $('#ib2-plugin-url').val());
		}

		if ( type == 'slider' ) {
			$('#' + id + '-num').val(val);
		} else if ( type == 'color' ) {
			var inputColor = this.utils.inputTextColor(val);
			$('#' + id).css({
				backgroundColor: val,
				color: inputColor
			});
		} else if ( type == 'icon' ) {
			$('#' + id).parent().find('.icon-preview').find('i').attr('class', val);
			if ( $('#' + id).parent().find('.btn-remove-icon').length && val !== '' ) {
				$('#' + id).parent().find('.btn-remove-icon').show();
			} else {
				$('#' + id).parent().find('.btn-remove-icon').hide();
			}
		} else if ( type == 'upload' ) {
			if ( val == '' ) {
				$('#' + id).parent().find('.remove-media').addClass('hidden');
				$('#' + id).closest('.settings-fields').find('.image-preview').addClass('hidden');
			} else {
				$('#' + id).parent().find('.remove-media').removeClass('hidden');
				$('#' + id).closest('.settings-fields').find('.image-preview').removeClass('hidden');
				$('#' + id).closest('.settings-fields').find('.image-preview').find('img').attr('src', val);
			}
		} else if ( type == 'bgpos' ) {
			var parent = $('#' + id).parent().find('.image-preview'),
				img = parent.find('img'),
				ball = parent.find('.pos-ball');

			if ( val == '0% 0%' ) {
				ball.css({
					top: '-10px',
					left: '-10px'
				});
			} else {
				var imgWidth = img.innerWidth(),
					imgHeight = img.innerHeight(),
					pct = val.replace(/%/gi, '').split(" ");

				ball.css({
					top: (Math.round(imgHeight * pct[1] / 100) - 10) + 'px',
					left: (Math.round(imgWidth * pct[0] / 100) - 10) + 'px'
				});
			}
		}

		if ( typeof tagName !== "undefined" && tagName.toLowerCase() == 'select' ) {
			var reveal = elem.attr('data-reveal');
			if ( typeof reveal !== "undefined" && reveal != 'none' && reveal !== null ) {
				elem.closest('.settings-fields').find('.' + reveal).addClass('select-hidden');
				elem.closest('.settings-fields').find('.' + val + '-' + reveal).removeClass('select-hidden');
			}

			if ( elem.attr('id').indexOf('popup-edit') > -1 ) {
				if ( val == 'none' || val == '' ) {
					elem.closest('.settings-fields').find('.edit-popup').addClass('hidden');
				} else {
					elem.closest('.settings-fields').find('.edit-popup').removeClass('hidden');
				}
			}
		}
	},

	loadElementIcons: function() {
		var sectionIdx = this.frameBody.find('.element-type-section').attr('data-idx') || 1;
		var section = '<div class="section-item" title="Section" data-el="section" data-idx="' + sectionIdx + '" data-toggle="tooltip" data-placement="right">';
			section += 'Section';
			section += '</div>';
		this.frameBody.find('#elements-bar > .elements-icons').append(section);
		this.elements.forEach(function(element, i){
			var type = element.hasOwnProperty('alias') ? element.alias : element.type,
				classes = (type != 'wsection' && type != 'page' && type != 'section') ? 'element-item' : 'hidden';

			if ( type == 'wbox' ) {
				classes = "element-section";
			}
			var html = '<div class="' + classes + ' element-type-' + type + '" title="' + element.title + '" data-el="' + type + '" data-idx="' + i + '" data-toggle="tooltip" data-placement="right">';
					if ( element.hasOwnProperty('textIcon') && element.textIcon !== '' ) {
						html += element.textIcon;
					} else {
						html += '<i class="' + element.icon + '"></i>';
					}
				html += '</div>';
			this.frameBody.find('#elements-bar > .elements-icons').append(html);
		}, this);
	},

	sortable: function(selector) {
		var self = this,
			cache = this.cache,
			iframeDom = this.frameBody;

		window.FrameWindow.$(".element-item").draggable({
			helper: "clone",
			iframeFix: true,
			connectToSortable: ".ib2-section-content",
			start: function(event, ui) {
				cache.isAdding = true;
				cache.isDragging = true;
			}
		});

		window.FrameWindow.$(".element-section,.section-item").draggable({
			helper: "clone",
			iframeFix: true,
			connectToSortable: "#screen-container",
			start: function(event, ui) {
				cache.isAdding = true;
				cache.isDragging = true;
			}
		});

		if ( cache.popupMode !== true && typeof selector === "undefined" || selector === null ) {
			selector = '.ib2-section-content';

			window.FrameWindow.$('#screen-container')
				.sortable({
					update: function(event, ui) {
						parent.$('body').trigger('sortableUpdated', [event, ui]);
					},
					stop: function(event, ui) {
						parent.$('body').trigger('sortableStopped', [event, ui]);
					},
					placeholder: 'sortable-big-line',
					items: ".ib2-section-el",
					cancel: ".ib2-slider-el,.ib2-wide-box",
				});
		}

		if ( !selector ) {
			selector = '.ib2-section-content';
		}

		window.FrameWindow.$(selector)
			.sortable({
				connectWith: ".ib2-section-content",
				forcePlaceholderSize: true,
				placeholder: 'sortable-line',
				items: ".ib2-content-el,.ib2-section-el",
				start: function(event, ui) {
					parent.$('body').trigger('sortableStarted', [event, ui]);
				},
				update: function(event, ui) {
					parent.$('body').trigger('sortableUpdated', [event, ui]);
				},
				stop: function(event, ui) {
					parent.$('body').trigger('sortableStopped', [event, ui]);
				}
			});

		window.FrameWindow.$('.ib2-wide-box')
			.draggable({
				containment: "parent",
				start: function(event, ui) {
					parent.$('body').trigger('sortableStarted', [event, ui]);
				},
				stop: function(event, ui) {
					parent.$('body').trigger('sortableStopped', [event, ui]);
				}
			});

		$('body').on('sortableUpdated', function(e, origEvent, ui){
			self.dropElement(origEvent, ui);
		
			self.saveHistory();
			self.savingHistory = true;
			setTimeout(function(){
				self.savingHistory = false;
			}, 2000);
		});

		$('body').on('sortableStarted', function(e, origEvent, ui){
			self.contentSortableStart(origEvent, ui);
		});

		$('body').on('sortableStopped', function(e, origEvent, ui){
			self.contentSortableStop(origEvent, ui);
		});
	},

	disableSortable: function() {
		var iframeDom = this.frameBody;
		iframeDom.find(".element-item").each(function(e){
			if ( $(this).draggable("instance") ) {
				$(this).draggable("destroy");
			}
		});

		iframeDom.find(".element-section").each(function(e){
			if ( $(this).draggable("instance") ) {
				$(this).draggable("destroy");
			}
		});

		if ( window.FrameWindow.$('#screen-container').sortable("instance") ) {
			window.FrameWindow.$('#screen-container').sortable("destroy");
		}

		window.FrameWindow.destroySortable();
	},

	contentSortableStart: function(event, ui) {
		var prime = window.primeObj,
			cache = prime.cache;

		cache.isDragging = true;

		// clean up
		$('.el-highlight').hide();
		$('.el-selected').hide();
		$('.is-selected').removeClass('is-selected');

		$('#editor-settings').find('.element-settings').remove();
	},

	contentSortableStop: function(event, ui) {
		var prime = window.primeObj,
			iframeDom = prime.frameBody,
			cache = prime.cache;

		cache.isDragging = false;
		
		$(prime.selector).trigger("sortStop", [$(ui.item).attr('id')]);

		cache.activeElement = null;
		cache.activeElementType = null;

		// remove is-selected class
		iframeDom.find('.is-selected').removeClass('is-selected');
	},

	dropElement: function(event, ui) { 
		var prime = window.primeObj,
			cache = prime.cache,
			target = event.target || null;

		cache.wboxTop = ui.offset.top;
		prime.addElement(target, ui.item);
	},

	addElement: function(target, elem, pos) { 
		var prime = window.primeObj,
			cache = prime.cache,
			gui = prime.gui,
			action = prime.action,
			iframeDom = prime.frameBody,
			settings = prime.settings,
			elements = prime.elements,
			type = elem.attr('data-el'),
			idx = elem.attr('data-idx'),
			newID = prime.generateID(8, 'ib2_el_');

		if ( target && cache.isAdding === true ) {
			var html = $('#ib2-element-' + type + '-html').html(),
				img = $('#ib2-img-src').attr('src'),
				vid = $('#ib2-vid-src').attr('src'),
				slide1 = $('#ib2-slide-1').attr('src'),
				slide2 = $('#ib2-slide-2').attr('src'),
				imgdir = $('#ib2-img-dir').val(),
				imgbtn = $('#ib2-submit-img').attr('src'),
				menuraw = $('#ib2-default-menu-html').html(),
				pagewidth = prime.pagewidth,
				colwidth = Math.round(pagewidth/2),
				offset = $('#default-time-zone').val();

			html = html.replace(new RegExp('{id}', 'gi'), newID); // replace id
			html = html.replace(new RegExp('{pagewidth}', 'gi'), pagewidth); // replace page width
			html = html.replace(new RegExp('{image}', 'gi'), img); // replace image
			html = html.replace(new RegExp('{video}', 'gi'), vid); // replace video placeholder image
			html = html.replace(new RegExp('{imgdir}', 'gi'), imgdir);
			html = html.replace(new RegExp('{imgbtn}', 'gi'), imgbtn);
			html = html.replace(new RegExp('{n}', 'gi'), 2);
			html = html.replace(new RegExp('{offset}', 'gi'), offset);
			html = html.replace(new RegExp('{colWidth}', 'gi'), colwidth);
			html = html.replace(new RegExp('{slide1}', 'gi'), slide1);
			html = html.replace(new RegExp('{slide2}', 'gi'), slide2);

			if ( type == 'wbox' ) {
				var boxTop = cache.wboxTop;
				if ( boxTop > 300 ) {
					boxTop = boxTop - 150;
				}
				html = html.replace(new RegExp('{top}', 'gi'), boxTop + 'px');
				cache.wboxTop = 0;
			}

			if ( type == 'date' ) {
				var today = moment().tz(offset).format('dddd, MMMM Do, YYYY');
				html = html.replace(new RegExp('{today}', 'gi'), today);
			}

			if ( type == 'menu' ) {
				var rawmenu = $(menuraw);
				rawmenu.find('ul').addClass('ib2-navi ib2-navi-plain');
				var menuhtml = rawmenu.html();
				html = html.replace(new RegExp('{menu}', 'gi'), menuhtml);
			}

			var helper = $(html);
			helper.find('.ib2-element').each(function(i){
				var $this = $(this),
					num = i+1,
					childID = prime.generateID(9, 'ib2_el_'),
					pattern = '{id' + num + '}',
					idRegex = new RegExp(pattern, 'gi');

				html = html.replace(idRegex, childID);
			});

			// replace html
			if ( pos == 'before' ) {
				$(html).insertBefore(target);
			} else if ( pos == 'after' ) {
				$(html).insertAfter(target);
			} else if ( pos == 'afterbegin' ) {
				target = target.hasClass('ib2-section-content') ? target : target.find('.ib2-section-content').eq(0);
				$(target).prepend(html);
			} else if ( pos == 'beforeend' ) {
				target = target.hasClass('ib2-section-content') ? target : target.find('.ib2-section-content').eq(0);
				$(target).append(html);
			} else {
				elem.replaceWith(html);
			}
			
			if ( type == 'countdown' ) {
				prime.updateSingleCountdown(iframeDom.find('#' + newID));
			}

			// save element data
			prime.newElementData(newID, type);

			if ( iframeDom.find('#' + newID).find('.ib2-element').length ) {
				iframeDom.find('#' + newID).find('.ib2-element').each(function(i){
					var $this = $(this),
						childID = $this.attr('id'),
						childType = $this.attr('data-el');

					prime.newElementData(childID, childType);
				});
			}

			if ( type == 'menu' ) {
				var navSetting = {
					'style': 'plain',
					'color': '#444444',
					'color@hover': '#333333',
					'backgroundColor' : 'transparent',
					'backgroundColor@hover': 'transparent',
					'sub-color': '#a7a7a7',
					'sub-color@hover': '#666666',
					'sub-backgroundColor': '#FFFFFF',
					'sub-backgroundColor@hover': '#E5E5E5'
				};
				cache.contents['menu'][newID]['settings']['menustyle'] = navSetting;
			}

			if ( prime.sortableReload ) {
				prime.disableSortable();
			}

			prime.timeoutHandler = setTimeout(function(){
				if ( prime.sortableReload ) {
					prime.sortable();
				}

				prime.sortableReload = false;
				cache.isAdding = false;
				if ( cache.isDragging === true ) {
					cache.isDragging = false;
				}

				cache.activeElement = null;
				cache.activeElementType = null;

				// remove is-selected class
				iframeDom.find('.is-selected').removeClass('is-selected');
				$('.el-selected > .el-resize-handlers').html('');
				$('.el-selected').hide();
				$('.el-highlight').hide();
				
				// close element settings
				// help close picker
				if ( $('.prime-colors').is(":visible") ) {
					prime.picker.pickerClose();
				}

				$('#editor-settings').find('.element-settings').remove();

				if ( cache.popupMode ) {
					var popupId = ( cache.sliderEdit ? cache.sliderEdit : cache.popupId );
					action.centerPopupEdit(iframeDom, popupId, cache.sliderEdit);
				}

				// re-calculate layers
				prime.saveLayers();

				clearTimeout(prime.timeoutHandler);
			}, 100);
		}

		// resize selector
		gui.resizeSelector();
		// resize backgrounds
		gui.resizeBackgrounds();
	},

	newElementData: function(newID, type) {
		var prime = window.primeObj,
			cache = prime.cache,
			iframeDom = prime.frameBody,
			element = prime.getElementData(type, prime.getElementAlias(iframeDom.find('#' + newID)));

		type = ( type == 'title' ? 'text' : type );

		// save settings
		if ( element && iframeDom.find('#' + newID).length ) {
			if ( element.hasOwnProperty('onDropped') && typeof element['onDropped'] === "function" ) {
				element['onDropped'](newID);
				// re-save custom styles
				prime.getCustomStyles();
			}

			var newSettings = prime.getSettingsData(type, newID),
				canHaveChild = ( !iframeDom.find('#' + newID).find(".ib2-section-content").length ) ? false : true,
				newHtml = ( iframeDom.find('#' + newID).find(' > .el-content').length ? iframeDom.find('#' + newID).find(' > .el-content').html() : iframeDom.find('#' + newID).html() ),
				width = iframeDom.find('#' + newID).find('.is-element').eq(0).length ? iframeDom.find('#' + newID).find('.is-element').eq(0).css('width') : iframeDom.find('#' + newID).css('width'),
				height = iframeDom.find('#' + newID).find('.is-element').eq(0).length ? iframeDom.find('#' + newID).find('.is-element').eq(0).css('height') : iframeDom.find('#' + newID).css('height'),
				marginTop = type == 'image' ? iframeDom.find('#' + newID).find('img').css('marginTop') : '0';
		
			// find correct dimension
			if ( element.hasOwnProperty('targetWidth') ) {
				var elem = prime.getSettingTarget(newID, element.targetWidth[0]);
				width = elem.css('width');
			}
		
			// adjust height
			if ( element.hasOwnProperty('targetHeight') ) {
				var elem = prime.getSettingTarget(newID, element.targetHeight[0]);
				height = elem.css('height');
			}

			var popName = null;
			if ( newID.indexOf('-popup') > -1 ) {
				var popnum = parseInt(iframeDom.find('.ib2-popup').length);
				iframeDom.find('#' + newID).show();
				height = iframeDom.find('#' + newID).css('height');
				iframeDom.find('#' + newID).hide();
				popName = 'Pop Up #' + popnum;
			}

			width = parseInt(width.replace(/px/gi, ''));
			height = parseInt(height.replace(/px/gi, ''));
			marginTop = parseInt(marginTop.replace(/px/gi, ''));

			if ( typeof cache.contents[type] === "undefined" ) {
				cache.contents[type] = {};
			}

			if ( Array.isArray(cache.contents[type]) ) {
				cache.contents[type] = {...cache.contents[type]};
			}

			cache.contents[type][newID] = {
				id: newID,
				settings: newSettings,
				type: type,
				width: width,
				height: height,
				marginTop: marginTop
			};

			// form
			if ( type == 'form' ) {
				cache.contents[type][newID]['inputs'] = prime.readInputs(iframeDom.find('#' + newID + ' > .is-element'));
			}

			if ( typeof popName === "string" && this.cache.contents['section'][newID]['settings'].hasOwnProperty('popsetup') ) {
				this.cache.contents['section'][newID]['settings']['popsetup']['popname'] = popName;
			}

			if ( element.hasOwnProperty('sortable') && element.sortable === true ) {
				this.sortableReload = true;
			}

			//console.log(cache.contents);
		}
	},

	toggleModeFields: function(type, id, mode) {
		$('#' + type + '-' + id).find('.settings-mode').find('.active').removeClass('btn-success active').addClass('btn-default');
		// set active 
		$('#' + type + '-' + id).find('.settings-mode').find('.mode-' + mode).removeClass('btn-default').addClass('btn-success active');
		// hide all fields first
		$('#fields-' + type + '-' + id).find('.form-group').addClass('hidden');
		// reveal correct fields
		$('#fields-' + type + '-' + id).find('.mode-' + mode).removeClass('hidden');
	},

	saveLayers: function() {
		var containers = window.FrameDocument.querySelectorAll("#screen-container > .ib2-section-el, #ib2-bottom-slider > .ib2-section-el");
		if ( containers ) {
			this.cache.layers = this.getDeepLayers(containers);
		}
	},

	getDeepLayers: function(containers) {
		var layers = [];

		if ( containers === null || typeof containers === "undefined" || !containers.length ) return null;
		containers.forEach(function(container){
			var layer = this.getLayer(container);
			layers.push(layer);
		}, this);
		return layers;
	},

	getLayer: function(container) {
		var $this = $(container),
			childEl = this.getContentElement(container),
			column = 0,
			parent = ( $this.parent().attr('id') == 'screen-container' ? '#screen-container' : null ),
			inColumn = false,
			cols = 1,
			colsizes = [],
			type = $this.attr('data-el');
		
		if ( $this.parent().hasClass('ib2-section-content') ) {
			// get parent type
			var parentType = $this.parent('.ib2-section-content').closest('.ib2-element').attr('data-el');
			if ( parentType == 'wsection' ) {
				parent = ' > .el-content > .el-content-inner > .el-cols > .ib2-section-content';
			} else {
				parent = ' > .el-content > .ib2-section-content';
			}
		} else {
			parent = ' > .el-content';
		}

		if ( type == 'section' || type == 'columns' ) {
			cols = $this.find(' > .el-content > .ib2-section-content').length;
		} else if ( type == 'wsection' ) {
			cols = $this.find(' > .el-content > .el-content-inner > .el-cols > .ib2-section-content').length;
		}

		if ( $this.parent().hasClass('ib2-section-col') ) {
			column = $this.parent().index();
			inColumn = true;
		}

		if ( $this.hasClass('ib2-section-el') && cols > 1 ) {
			if ( type == 'section' || type == 'columns' ) {
				$this.find(' > .el-content > .ib2-section-content').each(function(){
					colsizes.push($(this).outerWidth());
				});
			} else if ( type == 'wsection' ) {
				$this.find(' > .el-content > .el-content-inner > .el-cols > .ib2-section-content').each(function(){
					colsizes.push($(this).outerWidth());
				});
			}
		}
		var layer = {
			id: $this.attr('id'),
			type: type,
			child: this.getDeepLayers(childEl),
			parent: parent,
			column: column,
			inColumn: inColumn,
			cols: cols,
			colsizes: colsizes,
		};
		return layer;
	},

	grabOldPageSettings: function() {
		if ( typeof ib2globalsettings !== "undefined" && ib2globalsettings !== null && ib2globalsettings !== false ) {
			var ib2pagesettings = ib2globalsettings['settings'];

			this.cache.contents['page'] = {};
			this.cache.contents['page']['screen-container'] = ib2pagesettings;

			var fontFamily = this.cache.contents['page']['screen-container']['settings']['pagetypo']['fontFamily'];
			fontFamily = fontFamily.replace(new RegExp('"', 'gi'), "'");
			fontFamily = fontFamily.replace(new RegExp(', ', 'g'), ',');
			this.cache.contents['page']['screen-container']['settings']['pagetypo']['fontFamily'] = fontFamily;

			var linkColor = this.cache.contents['page']['screen-container']['settings']['pagetypo']['color@active'];
			linkColor = ( linkColor.substr(0, 1) !== "#" && linkColor.substr(0, 3) !== "rgb" ) ? '#428bca' : linkColor;
			this.cache.contents['page']['screen-container']['settings']['pagetypo']['color@active'] = linkColor;
		}
	},

	objectify: function() {
		// get the global page settings from the old editor
		this.grabOldPageSettings();
		
		// start objectifying elements
		var containers = window.FrameDocument.querySelectorAll("#screen-container .ib2-element"), popnum = 0;
		if ( containers === null || typeof containers === "undefined" || !containers.length ) return null;
		var iframeDom = this.frameBody;
		containers.forEach(function(container){
			var $this = $(container),
				findChild = ( !$this.find(".ib2-element").length ) ? false : true,
				childEl = this.getContentElement(container),
				type = $this.attr('data-el'),
				elID = $this.attr('id'),
				element = this.getElementData(type, this.getElementAlias($this)),
				width = $this.find('.is-element').eq(0).length ? $this.find('.is-element').eq(0).css('width') : $this.css('width'),
				height = $this.find('.is-element').eq(0).length ? $this.find('.is-element').eq(0).css('height') : $this.css('height'),
				marginTop = type == 'image' ? $this.find('img').css('marginTop') : '0',
				saveAs = type;
			
			// find correct dimension
			if ( element && element.hasOwnProperty('targetWidth') ) {
				var elem = this.getSettingTarget($this.attr('id'), element.targetWidth[0]);
				width = elem.css('width');
			}

			// adjust height
			if ( element && element.hasOwnProperty('targetHeight') ) {
				var elem = this.getSettingTarget($this.attr('id'), element.targetHeight[0]);
				height = elem.css('height');
			}

			var popName = null;
			if ( elID.indexOf('-popup') > -1 ) {
				popnum++;
				iframeDom.find('#' + elID).show();
				height = $this.css('height');
				iframeDom.find('#' + elID).hide();
				popName = 'Pop Up #' + popnum;
			}

			width = parseInt(width.replace(/px/gi, ''));
			height = parseInt(height.replace(/px/gi, ''));
			marginTop = parseInt(marginTop.replace(/px/gi, ''));

			if ( !this.cache.contents.hasOwnProperty(saveAs) ) {
				this.cache.contents[saveAs] = {};
			}

			if ( Array.isArray(this.cache.contents[saveAs]) ) {
				this.cache.contents[saveAs] = {...this.cache.contents[saveAs]};
			}

			this.cache.contents[saveAs][elID] = {
				id: elID,
				type: type,
				width: width,
				height: height,
				marginTop: marginTop,
				settings: this.getSettingsData(type, elID),
				inputs: ( type == 'form' ) ? this.readInputs($this) : null,
			};

			if ( typeof popName === "string" && this.cache.contents[saveAs][elID]['settings'].hasOwnProperty('popsetup') ) {
				this.cache.contents[saveAs][elID]['settings']['popsetup']['popname'] = popName;
			}
		}, this);
	},

	readInputs: function(el) {
		var inputs = [], 
			lastradio = null,
			icons = ["mail", "user", "home", "phone", "mobile", "search", "clock"];

		// get inputs
		el.find('input, select, textarea').each(function(){
			var $input = $(this),
				tagname = $input.prop('tagName').toLowerCase(),
				name = $input.attr('name'),
				inputtype = tagname == 'input' ? $input.attr('type') : tagname,
				label = $input.closest('label').text();

			if ( inputtype == 'submit' || inputtype == 'button' || inputtype == 'image' ) {
				return true;
			}

			if ( lastradio !== null && name == lastradio ) {
				return true;
			}

			var input = {
				name: name,
				tag: tagname,
				type: inputtype,
				value: $input.val() || '',
				placeholder: $input.attr('placeholder') || '',
				label: label || '',
				icon: 'none',
				isEmail: inputtype == 'email' ? true : false,
				visible: ( $input.is(":visible") && inputtype != 'hidden' ? true : false ),
				required: $input.attr('required') ? true : false,
				checked: ( inputtype == 'checkbox' ) ? $input.attr('checked') : false,
				text: ( inputtype == 'checkbox' ) ? $input.parent().text() : '',
				options: {}
			}

			icons.forEach(function(icon){
				if ( $input.hasClass('field-' + icon + '-icon') ) {
					input.icon = icon;
				}
			});

			if ( tagname == 'select' ) {
				input.options = {};
				$input.find('option').each(function(){
					input.options[$(this).val()] = $(this).text();
				});
			}

			if ( inputtype == 'radio' && $input.parents('form').find("input[name='" + name + "']").length ) {
				input.options = {};
				$input.parents('form').find("input[name='" + name + "']").each(function(){
					var $radio = $(this);
					input.options[$radio.val()] = $radio.parent().text() || $radio.val();
					if ( $radio.is(":checked") ) {
						input.value = $radio.val();
					}
					lastradio = name;
				});
			}

			if ( input.placeholder == '' ) {
				if ( input.label !== '' ) {
					input.placeholder = input.label;
				} else if ( input.label == '' && input.name !== '' ) {
					var name = input.name.toLowerCase();
					if ( name.indexOf('name') > -1 ) {
						input.placeholder = 'Your Name';
					} else if ( name.indexOf('email') > -1 ) {
						input.placeholder = 'Your Email Address';
					} else if ( name.indexOf('street') > -1 || name.indexOf('address') > -1 ) {
						input.placeholder = 'Address';
					} else if ( name.indexOf('city') > -1 ) {
						input.placeholder = 'City';
					} else if ( name.indexOf('state') > -1 ) {
						input.placeholder = 'State';
					} else if ( name.indexOf('province') > -1 ) {
						input.placeholder = 'Province';
					} else if ( name.indexOf('Country') > -1 ) {
						input.placeholder = 'Country';
					} else if ( name.indexOf('phone') > -1 ) {
						input.placeholder = 'Phone';
					} else if ( name.indexOf('zip') > -1 || name.indexOf('postal') > -1 ) {
						input.placeholder = 'Zip/Postal Code';
					} else {
						if ( input.isEmail === true ) {
							input.placeholder = 'Your Email Address';
						}
					}
				}
			}

			inputs.push(input);
		});
		return inputs;
	},

	readHtml: function (el) {
		if ( el.find('.el-content').eq(0).length ) {
			return el.find('.el-content').eq(0).html();
		} else {
			return el.html();
		}
	},

	readSettings: function(elem, keys, name, funcName, setting) {
		var cache = this.cache, 
			settings = {},
			elID = ( elem.hasClass('ib2-element') ? elem.attr('id') : elem.closest('.ib2-element').attr('id') ); 

		settings[name] = {};
		if ( $.isArray(keys) && keys.length > 0 ) {
			keys.forEach(function(key){
			 	settings[name][key] = this[funcName](elem, key, setting);
			}, this);
		} else if ( keys.hasOwnProperty('single') ) {
			settings[name]['mode'] = cache.isAdding ? 'all' : 'single';
			settings[name]['all'] = this[funcName](elem, keys.all, setting);
			settings[name]['single'] = {};
			keys.single.forEach(function(key){
			 	settings[name]['single'][key] = this[funcName](elem, key, setting);
			}, this);
		}
		return settings;
	},

	readSetting: function(elem, key, funcName, setting) {
		return this[funcName](elem, key, setting);
	},

	cssVarMap: function(type, field, prop) {
		var cssvar = '',
			sub = prop.indexOf('sub-') > -1 ? 'sub-' : '',
			hover = prop.indexOf('@hover') > -1 ? '-hover' : '';

		switch (prop) {
			case 'backgroundColor@hover':
			case 'backgroundColor@active':
			case 'backgroundColor':
			case 'sub-backgroundColor':
			case 'sub-backgroundColor@hover':
				cssvar = '--' + sub + type + hover + '-bg';
				break;

			case 'color@active':
			case 'color@hover':
			case 'color':
			case 'sub-color@hover':
			case 'sub-color':
				cssvar = '--' + sub + type + hover + '-color';
				break;

			case 'textShadow@hover':
			case 'textShadow':
				cssvar = '--' + sub + type + hover + '-tshadow';
				break;
		}
		return cssvar;
	},

	settingFunctionMap: function(option) {
		var funcName = false,
			funcs = {
				file: 'getFile',
				prop: 'getStyle',
				attrib: 'getAttrib',
				textClass: 'getText',
				classname: 'getClass'
			},
			dataType = ['file', 'prop', 'attrib', 'textClass', 'classname'];

		dataType.forEach(function(key){
			if ( option.hasOwnProperty(key) && funcs.hasOwnProperty(key) ) {
				funcName = funcs[key];
			}
		});
		return funcName;
	},

	readType: function(elem, keys, name, setting) {
		var cache = this.cache, 
			settings = {}; 

		settings[name] = {};
		keys.forEach(function(key){
			if ( setting.hasOwnProperty('typeSettings') && setting['typeSettings'].hasOwnProperty(key) ) {
				var typeset = setting['typeSettings'][key];
				switch(typeset.checkType) {
					case 'class':
						for ( var k in typeset.mapType ) {
							if ( elem.parent().find('.' + k).length && elem.parent().find('.' + k).get(0).style.display != 'none' ) {
								settings[name][key] = typeset.mapType[k];
							}
						}
						break;
				}
			}
		 	
		}, this);
		return settings;
	},

	readDynamics: function(elem, keys, name, setting) {
		var cache = this.cache, 
			settings = {},
			iframeDom = this.frameBody,
			elID = ( elem.hasClass('ib2-element') ? elem.attr('id') : elem.closest('.ib2-element').attr('id') );

		settings[name] = {};
		keys.forEach(function(key){
			var fields = setting.fields,
				field;

			for ( var i = 0; i < fields.length; i++ ) {
				if ( fields[i].hasOwnProperty('dynamic')
					&& fields[i]['dynamic'].length > 0
					&& fields[i]['dynamic'][0]['dynkey'] == key
				) {
					field = fields[i]['dynamic'];
					break;
				}	
			}

			if ( field ) {
				settings[name][key] = [];
				var data = {};
				field.forEach(function(input){
					if ( input.dynkey == 'slides' ) {
						if ( typeof ib2globalsettings['carousel_data'] === "undefined" ) {
							ib2globalsettings['carousel_data'] = {};
						}

						if ( ib2globalsettings.hasOwnProperty('carousel_data') && ib2globalsettings['carousel_data'] === null ) {
							ib2globalsettings['carousel_data'] = {};
						}

						if ( ib2globalsettings.hasOwnProperty('carousel_data') && ib2globalsettings['carousel_data'].hasOwnProperty(elID) ) {
							var carousel = ib2globalsettings['carousel_data'][elID] || null;
							settings[name][key] = carousel;
						} else {
							var carousel = [];
							iframeDom.find('#' + elID).find('.carousel-inner > .item').each(function(){
								var imgslide = {}; 
								imgslide.desturl = ( $(this).find('a').length ? $(this).find('a').attr('href') : '' );
								imgslide.imageurl = $(this).find('img').attr('src');
								imgslide.title = ( $(this).find('.carousel-caption').length ? $(this).find('.carousel-caption').text() : '' );
								carousel.push(imgslide);
							});
							settings[name][key] = carousel;
						}
					} else if ( input.dynkey == 'tabs' ) {
						let navtabs = iframeDom.find('#' + elID).find('> ul.ib2-nav-tabs').find('.ib2-tab-title'),
							tabs = [];

						navtabs.each(function(){
							tabs.push({title: $(this).text()});
						});
						settings[name][key] = tabs;
					}
				});
			}
		}, this);

		return settings;
	},

	getSettingsData: function(type, id) {
		var iframeDom = this.frameBody,
			elementData = this.getElementData(type, this.getElementAlias(iframeDom.find('#' + id))),
			options, defSettings = this.settings,
			settings = {};

		if ( typeof elementData === "undefined" || !elementData.hasOwnProperty('options') ) {
			return null;
		}

		options = elementData.options;
		if ( options.length > 0 ) {
			options.forEach(function(option) {
				var name = option.name,
					target = option.target,
					elem = this.getSettingTarget(id, option.target),
					props = [], files = [], texts = [], attribs = [];

				if ( defSettings.hasOwnProperty(name) ) {
					settings[name] = {};
					settings[name]['pos'] = option.target;
					props = ( defSettings[name].hasOwnProperty('props') ? defSettings[name]['props'] : null );
					files = ( defSettings[name].hasOwnProperty('files') ? defSettings[name]['files'] : null );
					texts = ( defSettings[name].hasOwnProperty('texts') ? defSettings[name]['texts'] : null );
					attribs = ( defSettings[name].hasOwnProperty('attribs') ? defSettings[name]['attribs'] : null );
					classes = ( defSettings[name].hasOwnProperty('cssclasses') ? defSettings[name]['cssclasses'] : null ),
					types = ( defSettings[name].hasOwnProperty('types') ? defSettings[name]['types'] : null ),
					dynamics = ( defSettings[name].hasOwnProperty('dynamics') ? defSettings[name]['dynamics'] : null );

					if ( props !== null ) {
						$.extend(true, settings, this.readSettings(elem, props, name, 'getStyle', defSettings[name]));
					}

					if ( files !== null ) {
						$.extend(true, settings, this.readSettings(elem, files, name, 'getFile', defSettings[name]));
					}

					if ( texts !== null ) {
						$.extend(true, settings, this.readSettings(elem, texts, name, 'getText', defSettings[name]));
					}

					if ( attribs !== null ) {
						$.extend(true, settings, this.readSettings(elem, attribs, name, 'getAttrib', defSettings[name]));
					}

					if ( classes !== null ) {
						$.extend(true, settings, this.readSettings(elem, classes, name, 'getClass', defSettings[name]));
					}

					if ( types !== null ) {
						$.extend(true, settings, this.readType(elem, types, name, defSettings[name]));
					}

					if ( dynamics !== null ) {
						$.extend(true, settings, this.readDynamics(elem, dynamics, name, defSettings[name]));
					}
				}
			}, this);
		}
		//console.log(type, settings);
		return settings;
	},

	getSettingTarget: function(id, target) {
		var iframeDom = this.frameBody,
			elem = iframeDom.find('#' + id),
			cache = this.cache;

		if ( target == 'middle' ) {
			elem =  iframeDom.find('#' + id + " > .el-content");
		} else if ( target == 'inner' ) {
			elem =  iframeDom.find('#' + id + " > .el-content > .el-content-inner");
		} else if ( target == 'content' ) {
			elem =  ( iframeDom.find('#' + id + " > .el-content > .el-content-inner > .el-cols > .ib2-section-content").length ?
						iframeDom.find('#' + id + " > .el-content > .el-content-inner > .el-cols > .ib2-section-content") : 
						iframeDom.find('#' + id + " > .el-content > .ib2-section-content")
					);

			if ( cache.popupMode && iframeDom.find('#' + id).hasClass('ib2-popup') ) {
				elem = iframeDom.find('#' + id + " > .el-content > .ib2-pop-content");
			}
		} else if ( target == 'box' ) {
			elem = iframeDom.find('#' + id + " > .el-content > .ib2-box-wrapper");
		} else if ( target == 'boxcontent' ) {
			elem = iframeDom.find('#' + id + " > .el-content > .ib2-box-wrapper > .ib2-box-content");
		} else if ( target == 'button' ) {
			elem = iframeDom.find('#' + id + " > .el-content > .ib2-button");
		} else if ( target == 'submit' ) {
			elem = iframeDom.find('#' + id + " > .el-content > form .ib2-form-submit");
		} else if ( target == 'image' ) {
			elem = iframeDom.find('#' + id + " picture > img");
		} else if ( target == 'figure' ) {
			elem = iframeDom.find('#' + id + " > .el-content > figure");
		} else if ( target == 'form' ) {
			elem = iframeDom.find('#' + id + " > .el-content > form");
		} else if ( target == 'gdpr' ) {
			elem = iframeDom.find('#' + id + " > .el-content > form > .eu-checkbox");
		} else if ( target == 'fbopt' ) {
			elem = iframeDom.find('#' + id + " > .el-content > .ib2-facebook-optin");
		} else if ( target == 'field' ) {
			elem = iframeDom.find('#' + id + " > .el-content > form > .form-fields > .form-group > input");
		} else if ( target == 'nav' ) {
			elem = iframeDom.find('#' + id + " > nav > ul.ib2-navi");
		} else if ( target == 'line' ) {
			elem = iframeDom.find('#' + id + " > .el-content > .ib2-hline");
		} else if ( target == 'link' ) {
			elem = iframeDom.find('#' + id).find('a').length ? iframeDom.find('#' + id).find('a') : iframeDom.find('#' + id);
		} else if ( target == 'tabcontent' ) {
			elem = iframeDom.find('#' + id).find('.ib2-tab-content').eq(0);
		} else if ( target == 'tabpane' ) {
			elem = iframeDom.find('#' + id).find('.ib2-tab-pane');
		} else if ( target == 'optinslide' ) {
			elem = iframeDom.find('#' + id).find('>.ib2-section-slide');
		} else if ( target == 'progress' ) {
			elem = iframeDom.find('#' + id).find('> .progress > .progress-bar');
		} else if ( target == 'iframe' ) {
			elem = iframeDom.find('#' + id).find('iframe');
		} else if ( target == 'collapsehead' ) {
			elem = iframeDom.find('#' + id).find('.ib2-collapse-head');
		} else if ( target == 'collapsecontent' ) {
			elem = iframeDom.find('#' + id).find('.ib2-collapse-content');
		}else if ( target == 'icon' ) {
			elem = iframeDom.find('#' + id).find('i');
		}
		return elem;
	},

	getContentElement: function(parentEl) {
		// Wide Section with Columns/Section as direct child
		var element = this.findDirectChild(parentEl, '.el-content > .el-content-inner > .el-cols > .ib2-section-content > .ib2-element');
		if ( element.length ) {
			return element;
		}

		element = this.findDirectChild(parentEl, '.el-content > .ib2-section-content > .ib2-element');
		if ( element.length ) {
			return element;
		}

		element = this.findDirectChild(parentEl, '.el-content > .ib2-element');
		if ( element.length ) {
			return element;
		}

		return parentEl;
	},

    findDirectChild: function(node, selector) {
        removeId = false;

        if (node.getAttribute('id') === null) {
            node.setAttribute('id', 'ID_' + new Date().getTime())
            removeId = true;
        }

        var result = window.FrameDocument.querySelectorAll('#' + node.getAttribute('id') + ' > ' + selector)

        if (removeId)
            node.removeAttribute('id')

        return result;
    },

    baseKey: function() {
    	return $('#ib2-post-id').val() + '-' + $('#ib2-current-variation').val();
    },

    getLocalData: function() {
    	var self = this,
    		localkey = this.baseKey(),
    		data = localStorage.getItem(localkey + '-data'),
    		html = localStorage.getItem(localkey + '-html');

    	return {
    		'data': self.parseJson(data),
    		'html': html,
    	};
    },

    parseJson(str) {
	    try {
	    	var o = JSON.parse(str);
	    	if ( o && typeof o === "object") {
	    		return o;
	    	}
	    } catch (e) {
	        return false;
	    }
	    return false;
	},

    clearLocalData: function() {
    	var localkey = this.baseKey();
    		
    	localStorage.removeItem(localkey + '-data');
    	localStorage.getItem(localkey + '-html');
    },

    isAutosave: function() {
    	var prime = this,
    		options = prime.globals;

    	if ( options.hasOwnProperty('autosave') ) {
    		var autosave = parseInt(options.autosave);
    		return autosave;
    	}
    },

    saveHistory: function() {
    	var prime = this,
    		interact = prime.interact,
    		iframeDom = prime.frameBody;

    	if ( prime.cache.isTexting || prime.savingHistory ) {
    		return;
    	}

    	// save history
    	if ( typeof prime.history !== "undefined" ) {
	    	prime.history.save();
    	}

    	var autosave = prime.isAutosave();
    	// save cache to storage
    	if ( autosave ) {
    		var currentstamp = Math.floor(Date.now() / 1000),
				as_interval = currentstamp - this.lastautosave;
		
			if ( as_interval < 120 ) {
				return;
			}
		
			this.cache.isAutosave = true;
			interact.savePageData(interact);
		
			this.lastautosave = currentstamp;
	    }
    },

    loadTemplates: function(type, text, isChange) {
    	var action = this.action;
		$('.ib2-tmpls').find('.ib2-template-loader').show();
		$('.ib2-tmpls').find('.ib2-templates-content').html('');
		action.centerPopup('template-manager');

		isChange = isChange || false;
		var options = '',
			config = ib2templateconfig,
			subtype = $('#ib2-tmpl-subtype').val(),
			keyword = $('#ib2-tmpl-tags').val(),
			selected;

		if ( !isChange ) {
			options += '<option value=""> -- select page sub-type -- </option>';
			if ( config.hasOwnProperty(type) && config[type] !== null ) {
				var subtypes = config[type];
				for ( var key in subtypes ) {
					selected = ( key == subtype ? ' selected="selected"' : '' );
					options += '<option value="' + key + '"' + selected + '>' + subtypes[key] + '</option>';
				}
			}
			$('#ib2-tmpl-subtype').html(options);
		}

		$.post(ajaxurl, {
			action: 'ib2_show_templates',
			type: type,
			subtype: subtype,
			keyword: keyword
		}, function(response){
			$('.ib2-tmpls').find('.ib2-templates-content').html(response);
			$('.ib2-tmpls').find('.ib2-template-loader').hide();

			$('.ib2-templates-content').find('.ib2-tmpl-thumb > img').each(function(){
				$(this).attr('src', $(this).attr('data-src'));
			});
			action.centerPopup('template-manager');
		});
    },

    loadSavedElements: function() {
    	var action = this.action;
		$('.ib2-element-loader').show();
		$('.ib2-elements-content').html('');
		action.centerPopup('element-manager');
		$.get(ajaxurl, {
			action: 'ib2_show_saved_elements',
		}, function(response){
			$('.ib2-elements-content').html(response);
			$('.ib2-element-loader').hide();
			$('#elements-placeholder').html(response);
			$('.ib2-elements-content').find('.ib2-elem-thumb > img').each(function(){
				$(this).attr('src', $(this).attr('data-src'));
			});
			action.centerPopup('element-manager');
		});
    },

    /**
	 * Insert GUI HTML to an element
	 * @param  {String} $type
	 * @param  {HtmlElement} el
	 * @param  {String} id
	 * @param  {String} pos
	 * @return {Void}
	 */
	loadHtml: function($type, el, id, pos) {
		var html = this.getHtml($type);
		if ( typeof id === 'string' ) {
			html = html.replace('[[id]]', id);
		}

		html = html.replace('{imgdir}', $('#ib2-img-folder').val());
		if ( typeof pos === "undefined" ) {
			pos = 'beforeend';
		}
		if ( el && html !== '' ) {
			el.insertAdjacentHTML(pos, html);
		}
	},

	/**
	 * Get HTML markup of available GUI
	 * @param  {String} name
	 * @return {String}
	 */
	getHtml: function(name) {
		var html = this.html[name];
		if ( typeof html === "undefined" ) {
			return '';
		}
		return html;
	},

	getButtonGui: function(el, type, alias, hover) {
		return this.gui.getButtons(el, type, alias, hover);
	},
	getResizerGui: function(el, type, alias) {
		return this.gui.getResizer(el, type, alias);
	},
	getElementAlias: function(el) {
		if ( el.hasClass('ib2-title-el') ) {
			return 'title';
		}
		return false;
	},
	generateID: function(length, prefix) {
	    var chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXTZabcdefghiklmnopqrstuvwxyz'.split('');
	
	    if ( !length ) {
	        length = Math.floor(Math.random() * chars.length);
	    }
	
		if ( typeof prefix === "undefined" ) {
			prefix = '';
		}

	    var str = prefix + '';
	    for ( var i = 0; i < length; i++ ) {
	        str += chars[Math.floor(Math.random() * chars.length)];
	    }
	    
	    if ( this.frameBody.find('#' + str).length ) {
	    	str = this.generateID(length);
	    }
    	return str;
	},

	animationMap: function(effect) {
    	switch (effect) {
    		case 'blind':
    			effect = 'fadeInDown';
    			break;
    		case 'clip':
    			effect = 'fadeInUp';
    			break;
    		case 'pulsate':
    			effect = 'flash';
    			break;
    		case 'drop':
    			effect = 'bounceInDown';
    			break;
    		case 'slide':
    			effect = 'fadeInLeft';
    			break;
    		case 'scale':
    			effect = 'zoomIn';
    			break;
    		case 'explode':
    			effect = 'rubberBand';
    			break;
    		case 'highlight':
    			effect = 'lightSpeedIn';
    			break;
    		case 'puff':
    			effect = 'tada';
    			break;
    		case 'fold':
    			effect = 'flip';
    			break;
    	}
    	return effect;
    }
};

PrimeBuilder.defaultOptions = {
	element: {
		selector: 'is-element',
		active: 'is-selected'
	}
};