<?php

class Ib2ApiRequest {
	/**
	 * Base api uri
	 * 
     * @var string
     */
    protected $base_uri = '';

	/**
     * The path to the certificate file to use for https connections
     *
     * @var string  Defaults to .
     */
    protected $cert = null;

    /**
     * cURL options
     *
     * @var array
     */
    protected $curl_options = array();

    /**
     * Http headers
     *
     * @var array
     */
    protected $http_headers = array();

    /**
     * Construct
     * 
     * @param string $base_uri
     * @param string $cert
     *
     * @return Void
     */
    public function __construct($base_uri = '', $cert = null)
    {
        if (!extension_loaded('curl')) {
            throw new Exception('The PHP exention curl must be installed to use this library.');
        }

        $this->base_uri = $base_uri;
        $this->cert = $cert;
        if (!empty($this->cert) && !is_file($this->cert)) {
            return new WP_Error('nocert', 'The certificate file was not found');
        }
    }

    /**
     * 
     */
    public function setHeaders(array $http_headers) 
    {
        $this->http_headers = $http_headers;
    }

    /**
     * Call api resource
     *
     * @param string $endpoint Api resource endpoint
     * @param array  $parameters Array of parameters
     * @param string $http_method HTTP Method to use (POST, PUT, GET, HEAD, DELETE)
     * @param array  $http_headers HTTP headers
     * @return array
     */
    public function call($endpoint, $parameters = array(), $http_method = 'GET', array $http_headers = array())
    {
    	$url = $this->base_uri . $endpoint;
        if ( empty($http_headers) && !empty($this->http_headers) ) {
            $http_headers = $this->http_headers;
        }
    	return $this->executeRequest($url, $parameters, $http_method, $http_headers);
    }

    /**
     * Execute a request (with curl)
     *
     * @param string $url URL
     * @param mixed  $parameters Array of parameters
     * @param string $http_method HTTP Method
     * @param array  $http_headers HTTP Headers
     * @return mixed
     */
    private function executeRequest($url, $parameters = array(), $http_method = 'GET', array $http_headers = null)
    {
        // prepare parameters
        $http_method = strtoupper($http_method);
        if ( !empty($parameters) ) {
            if ( in_array($http_method, array('POST', 'PUT', 'PATCH')) ) {
                if ( $http_headers && isset($http_headers['Content-Type']) ) {
                    if ( $http_headers['Content-Type'] == 'application/json' ) {
                        $parameters = json_encode($parameters);
                    } else if ( $http_headers['Content-Type'] == 'application/x-www-form-urlencoded' ) {
                        $data = '';
                        foreach( $parameters as $key => $value ) $data .= urlencode($key) . '=' . urlencode($value) . '&';
                        $parameters = rtrim($data, '& ');
                    }
                } else {
                    $parameters = http_build_query($parameters, null, '&');
                }
            } else if ( in_array($http_method, array('HEAD', 'DELETE', 'GET')) ) {
                if ( is_array($parameters) ) {
                    $url .= '?' . http_build_query($parameters, null, '&');
                } else if ($parameters) {
                    $url .= '?' . $parameters;
                }
            }
        }

        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_CUSTOMREQUEST, $http_method);

        switch( $http_method ) {
            case 'POST':
                curl_setopt($ch, CURLOPT_POST, true);
                /* No break */
            case 'PUT':
			case 'PATCH':
                curl_setopt($ch, CURLOPT_POSTFIELDS, $parameters);
                break;
            case 'HEAD':
                curl_setopt($ch, CURLOPT_NOBODY, true);
                break;
            default:
                break;
        }

        if ( !empty($http_headers) ) {
            $headers = array();
            foreach($http_headers as $key => $parsed_value) {
                $headers[] = "{$key}: {$parsed_value}";
            }
            curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        }

        // https handling
        if (!empty($this->cert)) {
            curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, true);
            curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 2);
            curl_setopt($ch, CURLOPT_CAINFO, $this->cert);
        } else {
            // bypass ssl verification
            curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
            curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
        }

        $result = curl_exec($ch);
        $http_code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $content_type = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);

        $json_decode = null;
        if ($curl_error = curl_error($ch)) {
            return new WP_Error('curl_error', $curl_error);
        } else {
            $json_decode = @json_decode($result, true);
        }
        curl_close($ch);

        // check for http error
        $check = $this->checkForErrors($http_code, $result);
        if ( is_wp_error($check) ) {
            return $check;
        }

        return (null === $json_decode) ? $result : $json_decode;
    }

	private function checkForErrors($http_code, $result) {
		switch ($http_code) {
			case 200:
			case 201:
			case 204:
				$error = false;
				break;

			case 400:
				$error = new WP_Error('bad_request', '(400) Bad Request. Incorrect format.', $result);
				break;

			case 401:
				$error = new WP_Error('unauthorized', '(401) Unauthorized. Access Token key is invalid.', $result);
				break;

			case 403:
				$error = new WP_Error('forbidden', '(403) Forbidden. The app has insufficient rights to access the resources.', $result);
				break;

			case 404:
				$error = new WP_Error('not_found', '(404) Not Found. The API endpoint or the requested data is not exists.', $result);
				break;

			case 405:
				$error = new WP_Error('method_not_allowed', '(405) Method Not Allowed. HTTP method is not accepted.', $result);
				break;

			case 422:
				$error = new WP_Error('invalid', '(422) Unprocessable Entity. Missing or invalid request body.', $result);
				break;

			case 429:
				$error = new WP_Error('throttled', '(429) Too many requests. Too many requests hits the API too quickly.', $result);
				break;

			default:
				$error = new WP_Error('unknown', "({$http_code}) Unknown API Error", $result);
		}
        return $error;
	}
}